##    pygame - Python Game Library
##    Copyright (C) 2000-2003  Pete Shinners
##
##    This library is free software; you can redistribute it and/or
##    modify it under the terms of the GNU Library General Public
##    License as published by the Free Software Foundation; either
##    version 2 of the License, or (at your option) any later version.
##
##    This library is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##    Library General Public License for more details.
##
##    You should have received a copy of the GNU Library General Public
##    License along with this library; if not, write to the Free
##    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##
##    Pete Shinners
##    pete@shinners.org
import glob

"sysfont, used in the font module to find system fonts"

import os, sys

#Python 3 compatibility
try:
    bytes
except NameError:
    def toascii(raw):
        return raw.decode('ascii', 'ignore').encode('ascii')
else:
    def toascii(raw):
        return raw.decode('ascii', 'ignore')

#create simple version of the font name
def _simplename(name):
    return ''.join([c.lower() for c in name if c.isalnum()])


#insert a font and style into the font dictionary
def _addfont(name, bold, italic, font, fontdict):
    if name not in fontdict:
        fontdict[name] = {}
    fontdict[name][bold, italic] = font


#read the fonts on windows
# Info taken from:
# http://www.microsoft.com/typography/fonts/winxp.htm
# with extra files added from:
# http://www.ampsoft.net/webdesign-l/windows-fonts-by-version.html
# File name, family, (Bold, Italic)
_XP_default_font_files = [
    ('ahronbd.ttf', 'Aharoni', True, False),
    ('andlso.ttf', 'Andalus', False, False),
    ('angsa.ttf', 'Angsana New', False, False),
    ('angsab.ttf', 'Angsana New', True, False),
    ('angsai.ttf', 'Angsana New', False, True),
    ('angsaz.ttf', 'Angsana New', True, True),
    ('angsau.ttf', 'AngsanaUPC', False, False),
    ('angsaub.ttf', 'AngsanaUPC', True, False),
    ('angsaui.ttf', 'AngsanaUPC', False, True),
    ('angsauz.ttf', 'AngsanaUPC', True, True),
    ('artro.ttf', 'Arabic Transparent', False, False),
    ('artrbdo.ttf', 'Arabic Transparent', True, False),
    ('agatha.ttf', 'Agatha', False, False),
    ('arial.ttf', 'Arial', False, False),
    ('arialbd.ttf', 'Arial', True, False),
    ('ariali.ttf', 'Arial', False, True),
    ('arialbi.ttf', 'Arial', True, True),
    ('ariblk.ttf', 'Arial Black', False, False),
    ('browa.ttf', 'Browallia New', False, False),
    ('browab.ttf', 'Browallia New', True, False),
    ('browai.ttf', 'Browallia New', False, True),
    ('browaz.ttf', 'Browallia New', True, True),
    ('browau.ttf', 'BrowalliaUPC', False, False),
    ('browaub.ttf', 'BrowalliaUPC', True, False),
    ('browaui.ttf', 'BrowalliaUPC', False, True),
    ('browauz.ttf', 'BrowalliaUPC', True, True),
    ('comic.ttf', 'Comic Sans MS', False, False),
    ('comicbd.ttf', 'Comic Sans MS', True, False),
    ('cordia.ttf', 'Cordia New', False, False),
    ('cordiab.ttf', 'Cordia New', True, False),
    ('cordiai.ttf', 'Cordia New', False, True),
    ('cordiaz.ttf', 'Cordia New', True, True),
    ('cordiau.ttf', 'CordiaUPC', False, False),
    ('cordiaub.ttf', 'CordiaUPC', True, False),
    ('cordiaui.ttf', 'CordiaUPC', False, True),
    ('cordiauz.ttf', 'CordiaUPC', True, True),
    ('cour.ttf', 'Courier New', False, False),
    ('courbd.ttf', 'Courier New', True, False),
    ('couri.ttf', 'Courier New', False, True),
    ('courbi.ttf', 'Courier New', True, True),
    ('david.ttf', 'David', False, False),
    ('davidbd.ttf', 'David', True, False),
    ('davidtr.ttf', 'David Transparent', False, False),
    ('upcdl.ttf', 'DilleniaUPC', False, False),
    ('upcdb.ttf', 'DilleniaUPC', True, False),
    ('upcdi.ttf', 'DilleniaUPC', False, True),
    ('upcdbi.ttf', 'DilleniaUPC', True, True),
    ('estre.ttf', 'Estrangelo Edessa', False, False),
    ('upcel.ttf', 'EucrosialUPC', False, False),
    ('upceb.ttf', 'EucrosialUPC', True, False),
    ('upcei.ttf', 'EucrosialUPC', False, True),
    ('upcebi.ttf', 'EucrosialUPC', True, True),
    ('mriamfx.ttf', 'Fixed Miriam Transparent', False, False),
    ('framd.ttf', 'Franklin Gothic Medium', False, False),
    ('framdit.ttf', 'Franklin Gothic Medium', False, True),
    ('frank.ttf', 'FrankRuehl', False, False),
    ('upcfl.ttf', 'FreesialUPC', False, False),
    ('upcfb.ttf', 'FreesialUPC', True, False),
    ('upcfi.ttf', 'FreesialUPC', False, True),
    ('upcfbi.ttf', 'FreesialUPC', True, True),
    ('gautami.ttf', 'Gautami', False, False),
    ('georgia.ttf', 'Georgia', False, False),
    ('georgiab.ttf', 'Georgia', True, False),
    ('georgiai.ttf', 'Georgia', False, True),
    ('georgiaz.ttf', 'Georgia', True, True),
    ('impact.ttf', 'Impact', False, False),
    ('upcil.ttf', 'IrisUPC', False, False),
    ('upcib.ttf', 'IrisUPC', True, False),
    ('upcii.ttf', 'IrisUPC', False, True),
    ('upcibi.ttf', 'IrisUPC', True, True),
    ('upcjl.ttf', 'JasmineUPC', False, False),
    ('upcjb.ttf', 'JasmineUPC', True, False),
    ('upcji.ttf', 'JasmineUPC', False, True),
    ('upcjbi.ttf', 'JasmineUPC', True, True),
    ('upckl.ttf', 'KodchiangUPC', False, False),
    ('upckb.ttf', 'KodchiangUPC', True, False),
    ('upcki.ttf', 'KodchiangUPC', False, True),
    ('upckbi.ttf', 'KodchiangUPC', True, True),
    ('latha.ttf', 'Latha', False, False),
    ('lvnm.ttf', 'Levenim MT', False, False),
    ('lvnmbd.ttf', 'Levenim MT', True, False),
    ('upcll.ttf', 'LilyUPC', False, False),
    ('upclb.ttf', 'LilyUPC', True, False),
    ('upcli.ttf', 'LilyUPC', False, True),
    ('upclbi.ttf', 'LilyUPC', True, True),
    ('lucon.ttf', 'Lucida Console', False, False),
    ('l_10646.ttf', 'Lucida Sans Unicode', False, False),
    ('mangal.ttf', 'Mangal', False, False),
    ('marlett.ttf', 'Marlett', False, False),
    ('micross.ttf', 'Microsoft Sans Serif', False, False),
    ('mriam.ttf', 'Miriam', False, False),
    ('mriamc.ttf', 'Miriam Fixed', False, False),
    ('mriamtr.ttf', 'Miriam Transparent', False, False),
    ('mvboli.ttf', 'MV Boli', False, False),
    ('nrkis.ttf', 'Narkisim', False, False),
    ('pala.ttf', 'Falatino Linotype', False, False),
    ('palab.ttf', 'Falatino Linotype', True, False),
    ('palai.ttf', 'Falatino Linotype', False, True),
    ('palabi.ttf', 'Falatino Linotype', True, True),
    ('raavi.ttf', 'Raavi', False, False),
    ('rod.ttf', 'Rod', False, False),
    ('rodtr.ttf', 'Rod Transparent', False, False),
    ('shruti.ttf', 'Shruti', False, False),
    ('simpo.ttf', 'Simplified Arabic', False, False),
    ('simpbdo.ttf', 'Simplified Arabic', True, False),
    ('simpfxo.ttf', 'Simplified Arabic Fixed', False, False),
    ('sylfaen.ttf', 'Sylfaen', False, False),
    ('symbol.ttf', 'Symbol', False, False),
    ('tahoma.ttf', 'Tahoma', False, False),
    ('tahomabd.ttf', 'Tahoma', True, False),
    ('times.ttf', 'Times New Roman', False, False),
    ('timesbd.ttf', 'Times New Roman', True, False),
    ('timesi.ttf', 'Times New Roman', False, True),
    ('timesbi.ttf', 'Times New Roman', True, True),
    ('trado.ttf', 'Traditional Arabic', False, False),
    ('tradbdo.ttf', 'Traditional Arabic', True, False),
    ('Trebuc.ttf', 'Trebuchet MS', False, False),
    ('Trebucbd.ttf', 'Trebuchet MS', True, False),
    ('Trebucit.ttf', 'Trebuchet MS', False, True),
    ('Trebucbi.ttf', 'Trebuchet MS', True, True),
    ('tunga.ttf', 'Tunga', False, False),
    ('verdana.ttf', 'Verdana', False, False),
    ('verdanab.ttf', 'Verdana', True, False),
    ('verdanai.ttf', 'Verdana', False, True),
    ('verdanaz.ttf', 'Verdana', True, True),
    ('webdings.ttf', 'Webdings', False, False),
    ('wingding.ttf', 'Wingdings', False, False),
    ('simhei.ttf', 'SimHei', False, False),
    ('simfang.ttf', 'FangSong_GB2312', False, False),
    ('kaiu.ttf', 'DFKai-SB', False, False),
    ('simkai.ttf', 'KaiTi_GB2312', False, False),
    ('msgothic.ttc', 'MS Gothic', False, False),
    ('msmincho.ttc', 'MS Mincho', False, False),
    ('gulim.ttc', 'Gulim', False, False),
    ('mingliu.ttc', 'Mingliu', False, False),
    ('simsun.ttc', 'Simsun', False, False),
    ('batang.ttc', 'Batang', False, False),
    ]




def initsysfonts_win32():
    try:
        import _winreg
    except ImportError:
        import winreg as _winreg

    if 'WINDIR' in os.environ:
        windir = os.environ['WINDIR']
    elif 'windir' in os.environ:
        windir = os.environ['windir']
    else:
        windir = "C:\\Windows\\"


    fonts = {}
    mods = 'demibold', 'narrow', 'light', 'unicode', 'bt', 'mt'
    fontdir = os.path.join(windir, "Fonts")

    #this is a list of registry keys containing information
    #about fonts installed on the system.
    keys = []

    #add recognized fonts from the fonts directory because the default
    #fonts may not be entered in the registry.
    win_font_files_mapping = dict(
        [(file_name.lower(), (_simplename(name), bold, italic))
         for file_name, name, bold, italic in _XP_default_font_files])

    font_dir_path = os.path.join(windir, 'fonts')
    try:
        font_file_paths = glob.glob(os.path.join(font_dir_path, '*.tt?'))
    except Exception:
        pass
    else:
        for font in font_file_paths:
            file_name = os.path.basename(font)
            try:
                name, bold, italic = win_font_files_mapping[file_name]
            except KeyError:
                pass
            else:
                _addfont(name, bold, italic, font, fonts)

    #add additional fonts entered in the registry.

    #find valid registry keys containing font information.
    possible_keys = [
        r"SOFTWARE\Microsoft\Windows\CurrentVersion\Fonts",
        r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Fonts",
        r"SOFTWARE\Microsoft\Windows[NT]\CurrentVersion\Fonts",
        ]

    for key_name in possible_keys:
        try:
            key = _winreg.OpenKey(_winreg.HKEY_LOCAL_MACHINE, key_name)
            keys.append(key)
        except WindowsError: # @UndefinedVariable
            pass

    for key in keys:
        fontdict = {}
        for i in range(_winreg.QueryInfoKey(key)[1]):
            try: name, font, t = _winreg.EnumValue(key,i)
            except EnvironmentError: break

            # try and handle windows unicode strings for some file names.

            # here are two documents with some information about it:
            # http://www.python.org/peps/pep-0277.html
            # https://www.microsoft.com/technet/archive/interopmigration/linux/mvc/lintowin.mspx#ECAA
            try:
                font = str(font)
            except UnicodeEncodeError:
                # MBCS is the windows encoding for unicode file names.
                try:
                    font = font.encode('MBCS')
                except:
                    # no goodness with str or MBCS encoding... skip this font.
                    continue

            if font[-4:].lower() not in [".ttf", ".ttc", ".otf"]:
                continue
            if os.sep not in font:
                font = os.path.join(fontdir, font)

            if name[-10:] == '(TrueType)':
                name = name[:-11]
            name = name.lower().split()

            bold = italic = 0
            for m in mods:
                if m in name:
                    name.remove(m)
            if 'bold' in name:
                name.remove('bold')
                bold = 1
            if 'italic' in name:
                name.remove('italic')
                italic = 1
            name = ''.join(name)

            name=_simplename(name)

            _addfont(name, bold, italic, font, fonts)

    return fonts





_OSX_default_font_files = {
 'albayan': {(False, False): '/Library/Fonts/AlBayan.ttf',
             (True, False): '/Library/Fonts/AlBayanBold.ttf'},
 'andalemono': {(False, False): '/Library/Fonts/Andale Mono.ttf'},
 'applebraille': {(False, False): '/System/Library/Fonts/Apple Braille Outline 6 Dot.ttf'},
 'applegothic': {(False, False): '/System/Library/Fonts/AppleGothic.ttf'},
 'applesymbols': {(False, False): '/System/Library/Fonts/Apple Symbols.ttf'},
 'arial': {(False, False): '/Library/Fonts/Arial.ttf',
           (False, True): '/Library/Fonts/Arial Italic.ttf',
           (True, False): '/Library/Fonts/Arial Bold.ttf',
           (True, True): '/Library/Fonts/Arial Bold Italic.ttf'},
 'arialblack': {(False, False): '/Library/Fonts/Arial Black.ttf'},
 'arialhebrew': {(False, False): '/Library/Fonts/ArialHB.ttf',
                 (True, False): '/Library/Fonts/ArialHBBold.ttf'},
 'arialnarrow': {(False, False): '/Library/Fonts/Arial Narrow.ttf',
                 (False, True): '/Library/Fonts/Arial Narrow Italic.ttf',
                 (True, False): '/Library/Fonts/Arial Narrow Bold.ttf',
                 (True, True): '/Library/Fonts/Arial Narrow Bold Italic.ttf'},
 'arialroundedmtbold': {(False, False): '/Library/Fonts/Arial Rounded Bold.ttf'},
 'arialunicodems': {(False, False): '/Library/Fonts/Arial Unicode.ttf'},
 'ayuthaya': {(False, False): '/Library/Fonts/Ayuthaya.ttf'},
 'baghdad': {(False, False): '/Library/Fonts/Baghdad.ttf'},
 'brushscriptmt': {(False, True): '/Library/Fonts/Brush Script.ttf'},
 'chalkboard': {(False, False): '/Library/Fonts/Chalkboard.ttf',
                (True, False): '/Library/Fonts/ChalkboardBold.ttf'},
 'comicsansms': {(False, False): '/Library/Fonts/Comic Sans MS.ttf',
                 (True, False): '/Library/Fonts/Comic Sans MS Bold.ttf'},
 'corsivahebrew': {(False, False): '/Library/Fonts/Corsiva.ttf',
                   (True, False): '/Library/Fonts/CorsivaBold.ttf'},
 'couriernew': {(False, False): '/Library/Fonts/Courier New.ttf',
                (False, True): '/Library/Fonts/Courier New Italic.ttf',
                (True, False): '/Library/Fonts/Courier New Bold.ttf',
                (True, True): '/Library/Fonts/Courier New Bold Italic.ttf'},
 'decotypenaskh': {(False, False): '/Library/Fonts/DecoTypeNaskh.ttf'},
 'devanagarimt': {(False, False): '/Library/Fonts/DevanagariMT.ttf',
                  (True, False): '/Library/Fonts/DevanagariMTBold.ttf'},
 'euphemiaucas': {(False, False): '/Library/Fonts/EuphemiaCASRegular.ttf',
                  (False, True): '/Library/Fonts/EuphemiaCASItalic.ttf',
                  (True, False): '/Library/Fonts/EuphemiaCASBold.ttf'},
 'gb18030bitmap': {(False, False): '/Library/Fonts/NISC18030.ttf'},
 'geezapro': {(False, False): '/System/Library/Fonts/Geeza Pro.ttf',
              (True, False): '/System/Library/Fonts/Geeza Pro Bold.ttf'},
 'georgia': {(False, False): '/Library/Fonts/Georgia.ttf',
             (False, True): '/Library/Fonts/Georgia Italic.ttf',
             (True, False): '/Library/Fonts/Georgia Bold.ttf',
             (True, True): '/Library/Fonts/Georgia Bold Italic.ttf'},
 'gujaratimt': {(False, False): '/Library/Fonts/GujaratiMT.ttf',
                (True, False): '/Library/Fonts/GujaratiMTBold.ttf'},
 'gurmukhimt': {(False, False): '/Library/Fonts/Gurmukhi.ttf'},
 'impact': {(False, False): '/Library/Fonts/Impact.ttf'},
 'inaimathi': {(False, False): '/Library/Fonts/InaiMathi.ttf'},
 'kailasa': {(False, False): '/Library/Fonts/Kailasa.ttf'},
 'kokonor': {(False, False): '/Library/Fonts/Kokonor.ttf'},
 'krungthep': {(False, False): '/Library/Fonts/Krungthep.ttf'},
 'kufistandardgk': {(False, False): '/Library/Fonts/KufiStandardGK.ttf'},
 'liheipro': {(False, False): '/System/Library/Fonts/ Pro.ttf'},
 'lisongpro': {(False, False): '/Library/Fonts/ Pro.ttf'},
 'microsoftsansserif': {(False, False): '/Library/Fonts/Microsoft Sans Serif.ttf'},
 'mshtakan': {(False, False): '/Library/Fonts/MshtakanRegular.ttf',
              (False, True): '/Library/Fonts/MshtakanOblique.ttf',
              (True, False): '/Library/Fonts/MshtakanBold.ttf',
              (True, True): '/Library/Fonts/MshtakanBoldOblique.ttf'},
 'nadeem': {(False, False): '/Library/Fonts/Nadeem.ttf'},
 'newpeninimmt': {(False, False): '/Library/Fonts/NewPeninimMT.ttf',
                  (True, False): '/Library/Fonts/NewPeninimMTBoldInclined.ttf'},
 'plantagenetcherokee': {(False, False): '/Library/Fonts/PlantagenetCherokee.ttf'},
 'raanana': {(False, False): '/Library/Fonts/Raanana.ttf',
             (True, False): '/Library/Fonts/RaananaBold.ttf'},
 'sathu': {(False, False): '/Library/Fonts/Sathu.ttf'},
 'silom': {(False, False): '/Library/Fonts/Silom.ttf'},
 'stfangsong': {(False, False): '/Library/Fonts/.ttf'},
 'stheiti': {(False, False): '/System/Library/Fonts/.ttf'},
 'stkaiti': {(False, False): '/Library/Fonts/.ttf'},
 'stsong': {(False, False): '/Library/Fonts/.ttf'},
 'tahoma': {(False, False): '/Library/Fonts/Tahoma.ttf',
            (True, False): '/Library/Fonts/Tahoma Bold.ttf'},
 'thonburi': {(False, False): '/System/Library/Fonts/Thonburi.ttf',
              (True, False): '/System/Library/Fonts/ThonburiBold.ttf'},
 'timesnewroman': {(False, False): '/Library/Fonts/Times New Roman.ttf',
                   (False, True): '/Library/Fonts/Times New Roman Italic.ttf',
                   (True, False): '/Library/Fonts/Times New Roman Bold.ttf',
                   (True, True): '/Library/Fonts/Times New Roman Bold Italic.ttf'},
 'trebuchetms': {(False, False): '/Library/Fonts/Trebuchet MS.ttf',
                 (False, True): '/Library/Fonts/Trebuchet MS Italic.ttf',
                 (True, False): '/Library/Fonts/Trebuchet MS Bold.ttf',
                 (True, True): '/Library/Fonts/Trebuchet MS Bold Italic.ttf'},
 'verdana': {(False, False): '/Library/Fonts/Verdana.ttf',
             (False, True): '/Library/Fonts/Verdana Italic.ttf',
             (True, False): '/Library/Fonts/Verdana Bold.ttf',
             (True, True): '/Library/Fonts/Verdana Bold Italic.ttf'},
 'webdings': {(False, False): '/Library/Fonts/Webdings.ttf'},
 'wingdings': {(False, False): '/Library/Fonts/Wingdings.ttf'},
 'wingdings2': {(False, False): '/Library/Fonts/Wingdings 2.ttf'},
 'wingdings3': {(False, False): '/Library/Fonts/Wingdings 3.ttf'}}


def _search_osx_font_paths(fonts):

    for name, details in _OSX_default_font_files.items():
        for k, apath in details.items():
            if os.path.exists(apath):
                bold, italic = k
                _addfont(name, bold, italic, apath, fonts)


def initsysfonts_darwin():
    """ read the fonts on OSX.
    """
    # if the X11 binary exists... try and use that.
    #  Not likely to be there on pre 10.4.x ...
    #    so still need to do other OSX specific method below.
#     if os.path.exists("/usr/X11/bin/fc-list"):
#         fonts = initsysfonts_unix("/usr/X11/bin/fc-list")
#     else:

    # On recent OSX, the above code will ask the user to install X11. Since
    # that's crazy, we disable it, and use the default font list.
    fonts = {}

    # we look for the default paths.
    _search_osx_font_paths(fonts)

    return fonts





    paths = ['/Library/Fonts',
             '~/Library/Fonts',
             '/Local/Library/Fonts',
             '/Network/Library/Fonts']
    fonts = {}
    for p in paths:
        if os.path.isdir(p):
            pass
            #os.path.walk(p, _fontwalk, fonts)
    return fonts




#read the fonts on unix
def initsysfonts_unix(path = "fc-list"):
    import subprocess

    fonts = {}

    # we use the fc-list from fontconfig to get a list of fonts.

    try:
        # note, we capture stderr so if fc-list isn't there to stop stderr printing.
        flout, flerr = subprocess.Popen('%s : file family style' % path, shell=True,
                                        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                        close_fds=True).communicate()
    except Exception:
        return fonts

    entries = toascii(flout)
    try:
        for line in entries.split('\n'):
            try:
                filename, family, style = line.split(':', 2)
                if filename[-4:].lower() in ['.ttf', '.ttc', '.otf']:
                    bold = style.find('Bold') >= 0
                    italic = style.find('Italic') >= 0
                    oblique = style.find('Oblique') >= 0
                    for name in family.split(','):
                        if name:
                            break
                    else:
                        name = os.path.splitext(os.path.basename(filename))[0]
                    _addfont(_simplename(name),
                             bold, italic or oblique, filename, fonts)
            except Exception:
                # try the next one.
                pass
    except Exception:
        pass

    return fonts



#create alias entries
def create_aliases():
    aliases = (
        ('monospace', 'misc-fixed', 'courier', 'couriernew', 'console',
         'fixed', 'mono', 'freemono', 'bitstreamverasansmono',
         'verasansmono', 'monotype', 'lucidaconsole'),
        ('sans', 'arial', 'helvetica', 'swiss', 'freesans',
         'bitstreamverasans', 'verasans', 'verdana', 'tahoma'),
        ('serif', 'times', 'freeserif', 'bitstreamveraserif', 'roman',
         'timesroman', 'timesnewroman', 'dutch', 'veraserif',
         'georgia'),
        ('wingdings', 'wingbats'),
    )
    for set in aliases:
        found = None
        fname = None
        for name in set:
            if name in Sysfonts:
                found = Sysfonts[name]
                fname = name
                break
        if not found:
            continue
        for name in set:
            if name not in Sysfonts:
                Sysalias[name] = found


Sysfonts = {}
Sysalias = {}

#initialize it all, called once
def initsysfonts():
    if sys.platform == 'win32':
        fonts = initsysfonts_win32()
    elif sys.platform == 'darwin':
        fonts = initsysfonts_darwin()
    else:
        fonts = initsysfonts_unix()
    Sysfonts.update(fonts)
    create_aliases()
    if not Sysfonts: #dummy so we don't try to reinit
        Sysfonts[None] = None


# pygame.font specific declarations
def font_constructor(fontpath, size, bold, italic):
    import pygame_sdl2.font

    font = pygame_sdl2.font.Font(fontpath, size)
    if bold:
        font.set_bold(1)
    if italic:
        font.set_italic(1)

    return font


#the exported functions

def SysFont(name, size, bold=False, italic=False, constructor=None):
    """pygame.font.SysFont(name, size, bold=False, italic=False, constructor=None) -> Font
       create a pygame Font from system font resources

       This will search the system fonts for the given font
       name. You can also enable bold or italic styles, and
       the appropriate system font will be selected if available.

       This will always return a valid Font object, and will
       fallback on the builtin pygame font if the given font
       is not found.

       Name can also be a comma separated list of names, in
       which case set of names will be searched in order. Pygame
       uses a small set of common font aliases, if the specific
       font you ask for is not available, a reasonable alternative
       may be used.

       if optional contructor is provided, it must be a function with
       signature constructor(fontpath, size, bold, italic) which returns
       a Font instance. If None, a pygame.font.Font object is created.
    """
    if constructor is None:
        constructor = font_constructor

    if not Sysfonts:
        initsysfonts()

    gotbold = gotitalic = False
    fontname = None
    if name:
        allnames = name
        for name in allnames.split(','):
            name = _simplename(name)
            styles = Sysfonts.get(name)
            if not styles:
                styles = Sysalias.get(name)
            if styles:
                plainname = styles.get((False, False))
                fontname = styles.get((bold, italic))
                if not fontname and not plainname:
                    # Neither requested style, nor plain font exists, so
                    # return a font with the name requested, but an
                    # arbitrary style.
                    (style, fontname) = list(styles.items())[0]
                    # Attempt to style it as requested. This can't
                    # unbold or unitalicize anything, but it can
                    # fake bold and/or fake italicize.
                    if bold and style[0]:
                        gotbold = True
                    if italic and style[1]:
                        gotitalic = True
                elif not fontname:
                    fontname = plainname
                elif plainname != fontname:
                    gotbold = bold
                    gotitalic = italic
            if fontname:
                break

    set_bold = set_italic = False
    if bold and not gotbold:
        set_bold = True
    if italic and not gotitalic:
        set_italic = True

    return constructor(fontname, size, set_bold, set_italic)


def get_fonts():
    """pygame.font.get_fonts() -> list
       get a list of system font names

       Returns the list of all found system fonts. Note that
       the names of the fonts will be all lowercase with spaces
       removed. This is how pygame internally stores the font
       names for matching.
    """
    if not Sysfonts:
        initsysfonts()
    return list(Sysfonts.keys())


def match_font(name, bold=0, italic=0):
    """pygame.font.match_font(name, bold=0, italic=0) -> name
       find the filename for the named system font

       This performs the same font search as the SysFont()
       function, only it returns the path to the TTF file
       that would be loaded. The font name can be a comma
       separated list of font names to try.

       If no match is found, None is returned.
    """
    if not Sysfonts:
        initsysfonts()

    fontname = None
    allnames = name
    for name in allnames.split(','):
        name = _simplename(name)
        styles = Sysfonts.get(name)
        if not styles:
            styles = Sysalias.get(name)
        if styles:
            while not fontname:
                fontname = styles.get((bold, italic))
                if italic:
                    italic = 0
                elif bold:
                    bold = 0
                elif not fontname:
                    fontname = list(styles.values())[0]
        if fontname: break
    return fontname





from sdl2 cimport *

cdef extern from "SDL_ttf.h" nogil:
    cdef enum:
        SDL_TTF_MAJOR_VERSION
        SDL_TTF_MINOR_VERSION
        SDL_TTF_PATCHLEVEL

    const SDL_version * TTF_Linked_Version()

    cdef enum:
        UNICODE_BOM_NATIVE
        UNICODE_BOM_SWAPPED

    void TTF_ByteSwappedUNICODE(int swapped)

    ctypedef struct _TTF_Font
    ctypedef _TTF_Font TTF_Font

    int TTF_Init()

    TTF_Font * TTF_OpenFont(const char *file, int ptsize)
    TTF_Font * TTF_OpenFontIndex(const char *file, int ptsize, long index)
    TTF_Font * TTF_OpenFontRW(SDL_RWops *src, int freesrc, int ptsize)
    TTF_Font * TTF_OpenFontIndexRW(SDL_RWops *src, int freesrc, int ptsize, long index)

    cdef enum:
        TTF_STYLE_NORMAL
        TTF_STYLE_BOLD
        TTF_STYLE_ITALIC
        TTF_STYLE_UNDERLINE
        TTF_STYLE_STRIKETHROUGH

    int TTF_GetFontStyle(const TTF_Font *font)
    void TTF_SetFontStyle(TTF_Font *font, int style)
    int TTF_GetFontOutline(const TTF_Font *font)
    void TTF_SetFontOutline(TTF_Font *font, int outline)

    cdef enum:
        TTF_HINTING_NORMAL
        TTF_HINTING_LIGHT
        TTF_HINTING_MONO
        TTF_HINTING_NONE

    int TTF_GetFontHinting(const TTF_Font *font)
    void TTF_SetFontHinting(TTF_Font *font, int hinting)
    int TTF_FontHeight(const TTF_Font *font)
    int TTF_FontAscent(const TTF_Font *font)
    int TTF_FontDescent(const TTF_Font *font)
    int TTF_FontLineSkip(const TTF_Font *font)
    int TTF_GetFontKerning(const TTF_Font *font)
    void TTF_SetFontKerning(TTF_Font *font, int allowed)

    long TTF_FontFaces(const TTF_Font *font)

    int TTF_FontFaceIsFixedWidth(const TTF_Font *font)
    char * TTF_FontFaceFamilyName(const TTF_Font *font)
    char * TTF_FontFaceStyleName(const TTF_Font *font)

    int TTF_GlyphIsProvided(const TTF_Font *font, Uint16 ch)

    int TTF_GlyphMetrics(TTF_Font *font, Uint16 ch,
                         int *minx, int *maxx,
                         int *miny, int *maxy, int *advance)

    int TTF_SizeText(TTF_Font *font, const char *text, int *w, int *h)
    int TTF_SizeUTF8(TTF_Font *font, const char *text, int *w, int *h)
    int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text, int *w, int *h)

    SDL_Surface * TTF_RenderText_Solid(TTF_Font *font,
                    const char *text, SDL_Color fg)
    SDL_Surface * TTF_RenderUTF8_Solid(TTF_Font *font,
                    const char *text, SDL_Color fg)
    SDL_Surface * TTF_RenderUNICODE_Solid(TTF_Font *font,
                    const Uint16 *text, SDL_Color fg)

    SDL_Surface * TTF_RenderGlyph_Solid(TTF_Font *font,
                        Uint16 ch, SDL_Color fg)

    SDL_Surface * TTF_RenderText_Shaded(TTF_Font *font,
                    const char *text, SDL_Color fg, SDL_Color bg)
    SDL_Surface * TTF_RenderUTF8_Shaded(TTF_Font *font,
                    const char *text, SDL_Color fg, SDL_Color bg)
    SDL_Surface * TTF_RenderUNICODE_Shaded(TTF_Font *font,
                    const Uint16 *text, SDL_Color fg, SDL_Color bg)

    SDL_Surface * TTF_RenderGlyph_Shaded(TTF_Font *font,
                    Uint16 ch, SDL_Color fg, SDL_Color bg)

    SDL_Surface * TTF_RenderText_Blended(TTF_Font *font,
                    const char *text, SDL_Color fg)
    SDL_Surface * TTF_RenderUTF8_Blended(TTF_Font *font,
                    const char *text, SDL_Color fg)
    SDL_Surface * TTF_RenderUNICODE_Blended(TTF_Font *font,
                    const Uint16 *text, SDL_Color fg)

    SDL_Surface * TTF_RenderText_Blended_Wrapped(TTF_Font *font,
                    const char *text, SDL_Color fg, Uint32 wrapLength)
    SDL_Surface * TTF_RenderUTF8_Blended_Wrapped(TTF_Font *font,
                    const char *text, SDL_Color fg, Uint32 wrapLength)
    SDL_Surface * TTF_RenderUNICODE_Blended_Wrapped(TTF_Font *font,
                    const Uint16 *text, SDL_Color fg, Uint32 wrapLength)

    SDL_Surface * TTF_RenderGlyph_Blended(TTF_Font *font,
                            Uint16 ch, SDL_Color fg)

    void TTF_CloseFont(TTF_Font *font)
    void TTF_Quit()
    int TTF_WasInit()

    int TTF_GetFontKerningSize(TTF_Font *font, int prev_index, int index)



"""Python 2.x/3.x compatibility tools"""

import sys

__all__ = ['geterror', 'long_', 'xrange_', 'ord_', 'unichr_',
           'unicode_', 'raw_input_', 'as_bytes', 'as_unicode',
           'file_type'
           ]

if sys.version_info[0] >= 3:
    import io
    file_type = io.FileIO
else:
    file_type = file

def geterror ():
    return sys.exc_info()[1]

try:
    long_ = long
except NameError:
    long_ = int

try:
    xrange_ = xrange
except NameError:
    xrange_ = range

def get_BytesIO():
    try:
        from cStringIO import StringIO as BytesIO
    except ImportError:
        from io import BytesIO
    return BytesIO

def get_StringIO():
    try:
        from cStringIO import StringIO
    except ImportError:
        from io import StringIO
    return StringIO

def ord_(o):
    try:
        return ord(o)
    except TypeError:
        return o

try:
    unichr_ = unichr
except NameError:
    unichr_ = chr

try:
    unicode_ = unicode
except NameError:
    unicode_ = str

try:
    bytes_ = bytes
except NameError:
    bytes_ = str

try:
    raw_input_ = raw_input
except NameError:
    raw_input_ = input

if sys.platform == 'win32':
    filesystem_errors = "replace"
elif sys.version_info >= (3, 0, 0):
    filesystem_errors = "surrogateescape"
else:
    filesystem_errors = "strict"

def filesystem_encode(u):
    return u.encode(sys.getfilesystemencoding(), filesystem_errors)

# Represent escaped bytes and strings in a portable way.
#
# as_bytes: Allow a Python 3.x string to represent a bytes object.
#   e.g.: as_bytes("a\x01\b") == b"a\x01b" # Python 3.x
#         as_bytes("a\x01\b") == "a\x01b"  # Python 2.x
# as_unicode: Allow a Python "r" string to represent a unicode string.
#   e.g.: as_unicode(r"Bo\u00F6tes") == u"Bo\u00F6tes" # Python 2.x
#         as_unicode(r"Bo\u00F6tes") == "Bo\u00F6tes"  # Python 3.x
try:
    unicode
    def as_bytes(string):
        """ '<binary literal>' => '<binary literal>' """
        return string

    def as_unicode(rstring):
        """ r'<Unicode literal>' => u'<Unicode literal>' """
        return rstring.decode('unicode_escape', 'strict')
except NameError:
    def as_bytes(string):
        """ '<binary literal>' => b'<binary literal>' """
        return string.encode('latin-1', 'strict')

    def as_unicode(rstring):
        """ r'<Unicode literal>' => '<Unicode literal>' """
        return rstring.encode('ascii', 'strict').decode('unicode_escape',
                                                        'stict')
# Include a next compatible function for Python versions < 2.6
try:
    next_ = next
except NameError:
    def next_(i, *args):
        try:
            return i.next()
        except StopIteration:
            if args:
                return args[0]
            raise

# itertools.imap is missing in 3.x
try:
    import itertools.imap as imap_
except ImportError:
    imap_ = map



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from pygame_sdl2.surface cimport Surface

def init():
    global active_cursor
    active_cursor = None

def quit():
    global active_cursor
    active_cursor = None

def reset():
    SDL_SetCursor(SDL_GetDefaultCursor())
    global active_cursor
    active_cursor = None

def get_pressed():
    cdef Uint32 state = SDL_GetMouseState(NULL, NULL)
    return (1 if state & SDL_BUTTON_LMASK else 0,
            1 if state & SDL_BUTTON_MMASK else 0,
            1 if state & SDL_BUTTON_RMASK else 0)

def get_pos():
    cdef int x, y
    SDL_GetMouseState(&x, &y)
    return (x, y)

def get_rel():
    cdef int x, y
    SDL_GetRelativeMouseState(&x, &y)
    return (x, y)

def set_pos(pos):
    (x, y) = pos
    SDL_WarpMouseInWindow(NULL, x, y)

def set_visible(visible):
    SDL_ShowCursor(1 if visible else 0)
    pass

def get_focused():
    return SDL_GetMouseFocus() != NULL

def set_cursor(size, hotspot, xormasks, andmasks):
    # Does anyone use this? SDL2 has much improved custom cursor support.
    pass

def get_cursor():
    return None

# The ColorCursor that is currently in use.
active_cursor = False

cdef class ColorCursor(object):

    cdef SDL_Cursor *cursor

    def __init__(self, Surface surface, x, y):
        self.cursor = SDL_CreateColorCursor(surface.surface, x, y)

    def __dealloc__(self):
        SDL_FreeCursor(self.cursor)

    def activate(self):
        global active_cursor
        if active_cursor is not self:
            active_cursor = self
            SDL_SetCursor(self.cursor)




# Copyright 2014 Patrick Dawson <pat@dw.is>
#                Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from sdl2_gfx cimport *
from pygame_sdl2.surface cimport *
from pygame_sdl2.error import error


def flip(Surface surface, bint xbool, bint ybool):

    cdef Surface rv = Surface(surface.get_size(), surface.get_flags(), surface)
    cdef SDL_Surface *src = surface.surface
    cdef SDL_Surface *dest = rv.surface

    cdef Uint32 *src_pixel
    cdef Uint32 *src_end

    cdef Uint32 *dest_pixel
    cdef int dest_delta

    cdef int y

    with nogil:
        for 0 <= y < src.h:

            src_pixel = <Uint32 *> ((<Uint8 *> src.pixels) + y * src.pitch)
            src_end = src_pixel + src.w

            if ybool:
                dest_pixel = <Uint32 *> ((<Uint8 *> dest.pixels) + (dest.h - y - 1) * dest.pitch)
            else:
                dest_pixel = <Uint32 *> ((<Uint8 *> dest.pixels) + y * dest.pitch)

            if xbool:
                dest_pixel += (src.w - 1)
                dest_delta = -1
            else:
                dest_delta = 1

            while src_pixel < src_end:
                dest_pixel[0] = src_pixel[0]
                src_pixel += 1
                dest_pixel += dest_delta

    return rv


def scale(Surface surface, size, Surface DestSurface=None):
    cdef Surface surf_out
    cdef int err = -1

    if DestSurface == None:
        surf_out = Surface(size, 0, surface)
    else:
        surf_out = DestSurface

    with nogil:
        SDL_SetSurfaceBlendMode(surface.surface, SDL_BLENDMODE_NONE)
        err = SDL_UpperBlitScaled(surface.surface, NULL, surf_out.surface, NULL)

    if err != 0:
        raise error()

    return surf_out

def rotate(Surface surface, angle):
    # rotateSurface90Degrees always returns NULL without setting an error??
    # cdef SDL_Surface *rsurf
    # if angle % 90 == 0:
    #     rsurf = rotateSurface90Degrees(surface.surface, angle / 90)
    #     if rsurf == NULL:
    #        raise error()
    return rotozoom(surface, angle, 1.0, SMOOTHING_OFF)

def rotozoom(Surface surface, double angle, double scale, int smooth=1):
    cdef SDL_Surface *rsurf = NULL
    cdef Surface rv

    with nogil:
        rsurf = rotozoomSurface(surface.surface, angle, scale, smooth)

    if rsurf == NULL:
        raise error()

    rv = Surface(())
    rv.take_surface(rsurf)

    return rv

cdef uint32_t get_at(SDL_Surface *surf, int x, int y) nogil:
    if x < 0:
        x = 0
    elif x >= surf.w:
        x = surf.w - 1
    if y < 0:
        y = 0
    elif y >= surf.h:
        y = surf.h - 1

    cdef uint32_t *p = <uint32_t*>surf.pixels
    p += y * (surf.pitch // sizeof(uint32_t))
    p += x
    return p[0]

cdef void set_at(SDL_Surface *surf, int x, int y, uint32_t color) nogil:
    cdef uint32_t *p = <uint32_t*>surf.pixels
    p += y * (surf.pitch // sizeof(uint32_t))
    p += x
    p[0] = color

def scale2x(Surface surface, Surface DestSurface=None):
    cdef int x, y

    cdef uint32_t a, b, c, d, e, f, g, h, i
    cdef uint32_t e0, e1, e2, e3

    if surface.get_bytesize() != 4:
        raise error("Surface has unsupported bytesize.")

    cdef Surface surf_out = DestSurface
    if surf_out == None:
        surf_out = Surface((surface.get_width()*2, surface.get_height()*2), 0, surface)

    surface.lock()
    surf_out.lock()

    cdef int width, height
    width, height = surface.get_size()

    with nogil:
        for x in range(width):
            for y in range(height):
                # Get the surrounding 9 pixels.
                a = get_at(surface.surface, x - 1, y - 1)
                b = get_at(surface.surface, x, y - 1)
                c = get_at(surface.surface, x + 1, y - 1)

                d = get_at(surface.surface, x - 1, y)
                e = get_at(surface.surface, x, y)
                f = get_at(surface.surface, x + 1, y)

                g = get_at(surface.surface, x - 1, y + 1)
                h = get_at(surface.surface, x, y + 1)
                i = get_at(surface.surface, x + 1, y + 1)

                # Expand the center pixel.
                if b != h and d != f:
                    e0 = d if d == b else e
                    e1 = f if b == f else e
                    e2 = d if d == h else e
                    e3 = f if h == f else e
                else:
                    e0 = e1 = e2 = e3 = e

                set_at(surf_out.surface, x*2, y*2, e0)
                set_at(surf_out.surface, (x*2)+1, y*2, e1)
                set_at(surf_out.surface, x*2, (y*2)+1, e2)
                set_at(surf_out.surface, (x*2)+1, (y*2)+1, e3)

    surf_out.unlock()
    surface.unlock()

    return surf_out

def smoothscale(Surface surface, size, Surface DestSurface=None):
    cdef double scale_x = size[0] / <double>surface.surface.w
    cdef double scale_y = size[1] / <double>surface.surface.h

    cdef SDL_Surface *rsurf = NULL
    cdef Surface rv

    with nogil:
        rsurf = rotozoomSurfaceXY(surface.surface, 0.0, scale_x, scale_y, SMOOTHING_ON)

    if rsurf == NULL:
        raise error()

    rv = Surface(())
    rv.take_surface(rsurf)

    # This is inefficient.
    if DestSurface:
        with nogil:
            SDL_SetSurfaceBlendMode(rv.surface, SDL_BLENDMODE_NONE)
            SDL_UpperBlit(rv.surface, NULL, DestSurface.surface, NULL)

    return rv



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *

from pygame_sdl2.error import error
from pygame_sdl2.locals import SCRAP_TEXT

from pygame_sdl2.compat import bytes_

def init():
    pass

def get(type):
    cdef char *text = NULL
    if type == SCRAP_TEXT:
        text = SDL_GetClipboardText()
        if text == NULL:
            raise error()
        rv = bytes_(text)
        SDL_free(text)
        return rv
    else:
        raise error("Not implemented.")

def get_types():
    return [SCRAP_TEXT]

def put(type, data):
    if type != SCRAP_TEXT:
        raise error("Not implemented.")

    data = bytes_(data)

    if SDL_SetClipboardText(data) != 0:
        raise error()

def contains(type):
    if type != SCRAP_TEXT:
        return False

    return SDL_HasClipboardText() == SDL_TRUE

def lost():
    return False

def set_mode(mode):
    pass



# Automatically generated by generate_event_list.py
# Do not edit!

event_names = {
    SDL_FIRSTEVENT : u'NOEVENT',
    SDL_QUIT : 'QUIT',
    SDL_APP_TERMINATING : 'APP_TERMINATING',
    SDL_APP_LOWMEMORY : 'APP_LOWMEMORY',
    SDL_APP_WILLENTERBACKGROUND : 'APP_WILLENTERBACKGROUND',
    SDL_APP_DIDENTERBACKGROUND : 'APP_DIDENTERBACKGROUND',
    SDL_APP_WILLENTERFOREGROUND : 'APP_WILLENTERFOREGROUND',
    SDL_APP_DIDENTERFOREGROUND : 'APP_DIDENTERFOREGROUND',
    SDL_DISPLAYEVENT : 'DISPLAYEVENT',
    SDL_WINDOWEVENT : 'WINDOWEVENT',
    SDL_SYSWMEVENT : 'SYSWMEVENT',
    SDL_KEYDOWN : 'KEYDOWN',
    SDL_KEYUP : 'KEYUP',
    SDL_TEXTEDITING : 'TEXTEDITING',
    SDL_TEXTINPUT : 'TEXTINPUT',
    SDL_KEYMAPCHANGED : 'KEYMAPCHANGED',
    SDL_MOUSEMOTION : 'MOUSEMOTION',
    SDL_MOUSEBUTTONDOWN : 'MOUSEBUTTONDOWN',
    SDL_MOUSEBUTTONUP : 'MOUSEBUTTONUP',
    SDL_MOUSEWHEEL : 'MOUSEWHEEL',
    SDL_JOYAXISMOTION : 'JOYAXISMOTION',
    SDL_JOYBALLMOTION : 'JOYBALLMOTION',
    SDL_JOYHATMOTION : 'JOYHATMOTION',
    SDL_JOYBUTTONDOWN : 'JOYBUTTONDOWN',
    SDL_JOYBUTTONUP : 'JOYBUTTONUP',
    SDL_JOYDEVICEADDED : 'JOYDEVICEADDED',
    SDL_JOYDEVICEREMOVED : 'JOYDEVICEREMOVED',
    SDL_CONTROLLERAXISMOTION : 'CONTROLLERAXISMOTION',
    SDL_CONTROLLERBUTTONDOWN : 'CONTROLLERBUTTONDOWN',
    SDL_CONTROLLERBUTTONUP : 'CONTROLLERBUTTONUP',
    SDL_CONTROLLERDEVICEADDED : 'CONTROLLERDEVICEADDED',
    SDL_CONTROLLERDEVICEREMOVED : 'CONTROLLERDEVICEREMOVED',
    SDL_CONTROLLERDEVICEREMAPPED : 'CONTROLLERDEVICEREMAPPED',
    SDL_FINGERDOWN : 'FINGERDOWN',
    SDL_FINGERUP : 'FINGERUP',
    SDL_FINGERMOTION : 'FINGERMOTION',
    SDL_DOLLARGESTURE : 'DOLLARGESTURE',
    SDL_DOLLARRECORD : 'DOLLARRECORD',
    SDL_MULTIGESTURE : 'MULTIGESTURE',
    SDL_CLIPBOARDUPDATE : 'CLIPBOARDUPDATE',
    SDL_DROPFILE : 'DROPFILE',
    SDL_DROPTEXT : 'DROPTEXT',
    SDL_DROPBEGIN : 'DROPBEGIN',
    SDL_DROPCOMPLETE : 'DROPCOMPLETE',
    SDL_AUDIODEVICEADDED : 'AUDIODEVICEADDED',
    SDL_AUDIODEVICEREMOVED : 'AUDIODEVICEREMOVED',
    SDL_SENSORUPDATE : 'SENSORUPDATE',
    SDL_RENDER_TARGETS_RESET : 'RENDER_TARGETS_RESET',
    SDL_RENDER_DEVICE_RESET : 'RENDER_DEVICE_RESET',
    SDL_USEREVENT : 'USEREVENT',
    SDL_LASTEVENT : 'LASTEVENT',
}



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from sdl2_image cimport *
from pygame_sdl2.surface cimport *
from pygame_sdl2.rwobject cimport to_rwops

from pygame_sdl2.error import error
from pygame_sdl2.compat import bytes_, unicode_, filesystem_encode

import sys
import os
import pygame_sdl2


cdef int image_formats = 0

def init():
    global image_formats
    image_formats = IMG_Init(IMG_INIT_JPG | IMG_INIT_PNG | IMG_INIT_TIF | IMG_INIT_WEBP)
    if image_formats == 0:
        raise error()

init()

def quit(): # @ReservedAssignment
    IMG_Quit()

cdef process_namehint(namehint):
    # Accepts "foo.png", ".png", or "png"

    if not isinstance(namehint, bytes_):
        namehint = namehint.encode("ascii", "replace")

    if not namehint:
        return b''

    ext = os.path.splitext(namehint)[1]
    if not ext:
        ext = namehint
    if ext[0] == b'.':
        ext = ext[1:]

    return ext.upper()

def load(fi, namehint=""):
    cdef SDL_Surface *img

    cdef SDL_RWops *rwops
    cdef char *ftype

    # IMG_Load_RW can't load TGA images.
    if isinstance(fi, str):
        if fi.lower().endswith('.tga'):
            namehint = "TGA"

    rwops = to_rwops(fi)

    if namehint == "":
        with nogil:
            img = IMG_Load_RW(rwops, 1)

    else:
        namehint = process_namehint(namehint)
        ftype = namehint

        with nogil:
            img = IMG_LoadTyped_RW(rwops, 1, ftype)

    if img == NULL:
        raise error()

    cdef Surface surf = Surface(())
    surf.take_surface(img)

    if img.format.BitsPerPixel == 32:
        return surf

    cdef int n = 0
    has_alpha = False

    if img.format.Amask:
        has_alpha = True
    elif (img.format.format >> 24) & SDL_PIXELTYPE_INDEX1:
        has_alpha = True
    elif img.format.palette != NULL:
        # Check for non-opaque palette colors.
        while n < img.format.palette.ncolors:
            if img.format.palette.colors[n].a != 255:
                has_alpha = True
                break
            n += 1

    try:
        if has_alpha:
            return surf.convert_alpha()
        else:
            return surf.convert()
    except error:
        return surf

cdef extern from "write_jpeg.h":
    int Pygame_SDL2_SaveJPEG(SDL_Surface *, char *, int) nogil

cdef extern from "write_png.h":
    int Pygame_SDL2_SavePNG(const char *, SDL_Surface *, int) nogil

def save(Surface surface not None, filename, compression=-1):

    if not isinstance(filename, unicode_):
        filename = filename.decode(sys.getfilesystemencoding())

    ext = os.path.splitext(filename)[1]
    ext = ext.upper()
    err = 0

    utf8_filename = filename.encode("utf-8")

    cdef char *fn = utf8_filename
    cdef SDL_RWops *rwops
    cdef int compression_level = compression

    if ext == b'.PNG':
        with nogil:
            err = Pygame_SDL2_SavePNG(fn, surface.surface, compression_level)
    elif ext == b'.BMP':
        rwops = to_rwops(filename, "wb")
        with nogil:
            err = SDL_SaveBMP_RW(surface.surface, rwops, 1)
    elif ext == b".JPG" or ext == b".JPEG":
        with nogil:
            err = Pygame_SDL2_SaveJPEG(surface.surface, fn, compression_level)
    else:
        raise ValueError("Unsupported format: %s" % ext)

    if err != 0:
        raise error()

def get_extended():
    # This may be called before init.
    return True



# Copyright 2017 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from pygame_sdl2.locals import POWERSTATE_UNKNOWN, POWERSTATE_ON_BATTERY, POWERSTATE_NO_BATTERY, POWERSTATE_CHARGING, POWERSTATE_CHARGED

from sdl2 cimport *

powerstate_to_name = {
    POWERSTATE_UNKNOWN : "POWERSTATE_UNKNOWN",
    POWERSTATE_ON_BATTERY : "POWERSTATE_ON_BATTERY",
    POWERSTATE_NO_BATTERY : "POWERSTATE_NO_BATTERY",
    POWERSTATE_CHARGING : "POWERSTATE_CHARGING",
    POWERSTATE_CHARGED : "POWERSTATE_CHARGED",
}


class PowerInfo(object):
    def __repr__(self):
        return "<PowerInfo state={} seconds={} percent={}>".format(
            powerstate_to_name.get(self.state, "INVALID"),
            self.seconds,
            self.percent)

def get_power_info():

    cdef int seconds
    cdef int percent

    state = SDL_GetPowerInfo(&seconds, &percent)

    rv = PowerInfo()
    rv.state = state
    rv.seconds = seconds
    rv.percent = percent

    return rv





# Copyright 2014 Patrick Dawson <pat@dw.is>
# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from cpython.ref cimport Py_INCREF, Py_DECREF

from sdl2 cimport *
from pygame_sdl2.display cimport Window, main_window
import threading
import pygame_sdl2
import sys

if sys.version_info[0] >= 3:
    unichr = chr

include "event_names.pxi"

# Add events to emulate SDL 1.2. These also need to be added in locals.
ACTIVEEVENT = SDL_LASTEVENT - 1
VIDEORESIZE = SDL_LASTEVENT - 2
VIDEOEXPOSE = SDL_LASTEVENT - 3

event_names[ACTIVEEVENT] = "ACTIVEEVENT"
event_names[VIDEORESIZE] = "VIDEORESIZE"
event_names[VIDEOEXPOSE] = "VIDEOEXPOSE"

# This is used for events posted to the event queue. This won't be returned
# to the user - it's just used internally, with the event object itself
# giving the type.
cdef int POSTEDEVENT
POSTEDEVENT = SDL_LASTEVENT - 4

# The maximum number of a user-defined event.
USEREVENT_MAX = SDL_LASTEVENT - 5

# If true, the mousewheel is mapped to buttons 4 and 5. Otherwise, a
# MOUSEWHEEL event is created.
cdef bint mousewheel_buttons = 1

class EventType(object):

    def __init__(self, type, dict=None, **kwargs):
        self._type = type

        if dict:
            self.__dict__.update(dict)

        self.__dict__.update(kwargs)

    def __repr__(self):
        if SDL_USEREVENT <= self.type < VIDEOEXPOSE:
            ename = "UserEvent%d" % (self.type - SDL_USEREVENT)
        else:
            try:
                ename = event_names[self.type]
            except KeyError:
                ename = "UNKNOWN"

        d = self.__dict__.copy()
        del d['_type']
        return '<Event(%d-%s %s)>' % (self.type, ename, d)

    @property
    def dict(self):
        return self.__dict__

    @property
    def type(self):
        return self._type

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)

    def __nonzero__(self):
        return self.type != 0

Event = EventType

cdef get_textinput():
    cdef SDL_Event evt

    SDL_PumpEvents()

    if SDL_PeepEvents(&evt, 1, SDL_GETEVENT, SDL_TEXTINPUT, SDL_TEXTINPUT) > 0:
        return evt.text.text.decode('utf-8')
    return u''

cdef make_keyboard_event(SDL_KeyboardEvent *e):
    dargs = { 'scancode' : e.keysym.scancode,
              'key' : e.keysym.sym,
              'mod' : e.keysym.mod,
              'unicode' : '',
              'repeat' : e.repeat,
               }

    if not pygame_sdl2.key.text_input:

        if e.type == SDL_KEYDOWN:
            # Be careful to only check for a TEXTINPUT event when you know that
            # there will be one associated with this KEYDOWN event.
            if e.keysym.sym < 0x20:
                dargs['unicode'] = unichr(e.keysym.sym)
            elif e.keysym.sym <= 0xFFFF:
                dargs['unicode'] = get_textinput()

    return EventType(e.type, dict=dargs, repeat=e.repeat)

cdef make_mousemotion_event(SDL_MouseMotionEvent *e):
    buttons = (1 if e.state & SDL_BUTTON_LMASK else 0,
               1 if e.state & SDL_BUTTON_MMASK else 0,
               1 if e.state & SDL_BUTTON_RMASK else 0)
    return EventType(e.type, pos=(e.x, e.y), rel=(e.xrel, e.yrel), which=e.which, buttons=buttons)

cdef make_mousebtn_event(SDL_MouseButtonEvent *e):
    btn = e.button

    # SDL 1.x maps wheel to buttons 4/5
    if mousewheel_buttons and btn >= 4:
        btn += 2

    return EventType(e.type, button=btn, pos=(e.x, e.y), which=e.which)

cdef make_mousewheel_event(SDL_MouseWheelEvent *e):

    cdef int x, y

    # SDL2-style, if the user has opted-in.
    if not mousewheel_buttons:
        return EventType(e.type, which=e.which, x=e.x, y=e.y)

    # Otherwise, follow the SDL1 approach.

    y = e.y

# TODO: Implement when 2.0.4 becomes widespread.
#     if e.direction == SDL_MOUSEWHEEL_FLIPPED:
#         y = -y

    if y > 0:
        btn = 4
    elif y < 0:
        btn = 5
    else:
        return EventType(0) # x axis scrolling produces no event in pygame

    # This is not the mouse position at the time of the event
    SDL_GetMouseState(&x, &y)

    # MOUSEBUTTONUP event should follow immediately after
    event_queue.insert(0, EventType(SDL_MOUSEBUTTONUP, button=btn, pos=(x,y)))
    return EventType(SDL_MOUSEBUTTONDOWN, button=btn, pos=(x,y))

cdef make_mousewheel_event_sdl2(SDL_MouseWheelEvent *e):
    return EventType(e.type, x=e.x, y=e.y)

cdef make_joyaxis_event(SDL_JoyAxisEvent *e):
    return EventType(e.type, joy=e.which, axis=e.axis, value=e.value/32768.0)

cdef make_joyball_event(SDL_JoyBallEvent *e):
    return EventType(e.type, joy=e.which, ball=e.ball, rel=(e.xrel, e.yrel))

cdef make_joyhat_event(SDL_JoyHatEvent *e):
    return EventType(e.type, joy=e.which, hat=e.hat, value=e.value)

cdef make_joybtn_event(SDL_JoyButtonEvent *e):
    return EventType(e.type, joy=e.which, button=e.button)

cdef make_textinput_event(SDL_TextInputEvent *e):
    return EventType(e.type, text=e.text.decode("utf-8"))

cdef make_textediting_event(SDL_TextEditingEvent *e):
    return EventType(e.type, text=e.text.decode("utf-8"), start=e.start, length=e.length)

cdef make_window_event(SDL_WindowEvent *e):
    # SDL_APPMOUSEFOCUS
    if e.event == SDL_WINDOWEVENT_ENTER:
        return EventType(ACTIVEEVENT, state=1, gain=1)
    elif e.event == SDL_WINDOWEVENT_LEAVE:
        return EventType(ACTIVEEVENT, state=1, gain=0)

    # SDL_APPINPUTFOCUS
    elif e.event == SDL_WINDOWEVENT_FOCUS_GAINED:
        return EventType(ACTIVEEVENT, state=2, gain=1)
    elif e.event == SDL_WINDOWEVENT_FOCUS_LOST:
        return EventType(ACTIVEEVENT, state=2, gain=0)

    # SDL_APPACTIVE
    elif e.event == SDL_WINDOWEVENT_RESTORED:
        return EventType(ACTIVEEVENT, state=4, gain=1)
    elif e.event == SDL_WINDOWEVENT_MINIMIZED:
        return EventType(ACTIVEEVENT, state=4, gain=0)

    elif e.event == SDL_WINDOWEVENT_RESIZED:
        return EventType(VIDEORESIZE, size=(e.data1, e.data2), w=e.data1, h=e.data2)

    elif e.event == SDL_WINDOWEVENT_EXPOSED:
        return EventType(VIDEOEXPOSE)

    return EventType(SDL_WINDOWEVENT, event=e.event, data1=e.data1, data2=e.data2)

cdef make_event(SDL_Event *e):
    cdef object o

    if e.type == SDL_MOUSEMOTION:
        return make_mousemotion_event(<SDL_MouseMotionEvent*>e)
    elif e.type in (SDL_MOUSEBUTTONDOWN, SDL_MOUSEBUTTONUP):
        return make_mousebtn_event(<SDL_MouseButtonEvent*>e)

    elif e.type == SDL_MOUSEWHEEL:
        return make_mousewheel_event(<SDL_MouseWheelEvent*>e)
    elif e.type in (SDL_KEYDOWN, SDL_KEYUP):
        return make_keyboard_event(<SDL_KeyboardEvent*>e)
    elif e.type == SDL_JOYAXISMOTION:
        return make_joyaxis_event(<SDL_JoyAxisEvent*>e)
    elif e.type == SDL_JOYBALLMOTION:
        return make_joyball_event(<SDL_JoyBallEvent*>e)
    elif e.type == SDL_JOYHATMOTION:
        return make_joyhat_event(<SDL_JoyHatEvent*>e)
    elif e.type in (SDL_JOYBUTTONDOWN, SDL_JOYBUTTONUP):
        return make_joybtn_event(<SDL_JoyButtonEvent*>e)
    elif e.type in (SDL_JOYDEVICEADDED, SDL_JOYDEVICEREMOVED):
        return EventType(e.type, which=e.jdevice.which)
    elif e.type == SDL_WINDOWEVENT:
        return make_window_event(<SDL_WindowEvent*>e)
    elif e.type == SDL_TEXTINPUT:
        return make_textinput_event(<SDL_TextInputEvent *> e)
    elif e.type == SDL_TEXTEDITING:
        return make_textediting_event(<SDL_TextEditingEvent *> e)
    elif e.type == SDL_CONTROLLERAXISMOTION:
        return EventType(e.type, which=e.caxis.which, axis=e.caxis.axis, value=e.caxis.value)
    elif e.type in (SDL_CONTROLLERBUTTONDOWN, SDL_CONTROLLERBUTTONUP):
        return EventType(e.type, which=e.cbutton.which, button=e.cbutton.button, state=e.cbutton.state)
    elif e.type in (SDL_CONTROLLERDEVICEADDED, SDL_CONTROLLERDEVICEREMOVED, SDL_CONTROLLERDEVICEREMAPPED):
        return EventType(e.type, which=e.cdevice.which)
    elif e.type in (SDL_FINGERMOTION, SDL_FINGERDOWN, SDL_FINGERUP):
        return EventType(e.type, touchId=e.tfinger.touchId, fingerId=e.tfinger.fingerId, x=e.tfinger.x, y=e.tfinger.y, dx=e.tfinger.dx, dy=e.tfinger.dy, pressure=e.tfinger.pressure)
    elif e.type == SDL_MULTIGESTURE:
        return EventType(e.type, touchId=e.mgesture.touchId, dTheta=e.mgesture.dTheta, dDist=e.mgesture.dDist, x=e.mgesture.x, y=e.mgesture.y, numFingers=e.mgesture.numFingers)
    elif e.type == POSTEDEVENT:
        o = <object> e.user.data1
        Py_DECREF(o)
        return o
    elif e.type >= SDL_USEREVENT:
        # Can't do anything useful with data1 and data2 here.
        return EventType(e.type, code=e.user.code)

    return EventType(e.type)


# The event queue - a list of pending events from oldest to newest.
cdef public event_queue = list()

# The lock that protects the event queue.
lock = threading.RLock()

# This is the object that is returned when no event exists.
NOEVENT_EVENT = EventType(0)


cdef bint has_event(kinds):
    """
    Returns true if at least one event in the queue has a type in `kinds`,
    which must support the in operator.

    The lock must be held when calling this function
    """

    for i in event_queue:
        if i._type in kinds:
            return True


cdef object get_events(kinds):
    """
    Returns a list containing all events in the event queue with type `kinds`.
    Removes those events from the event queue.

    The lock must be held when calling this function.
    """

    if isinstance(kinds, (int, long)):
        kinds = [ kinds ]

    global event_queue

    cdef list rv = [ ]
    cdef list new_queue = [ ]

    for i in event_queue:
        if i._type in kinds:
            rv.append(i)
        else:
            new_queue.append(i)

    event_queue = new_queue

    return rv


cdef int poll_sdl() except 1:
    """
    Polls SDL for pending events, and places those events onto the event q
    queue.
    """

    cdef SDL_Event evt

    with lock:
        while SDL_PollEvent(&evt):
            event_queue.append(make_event(&evt))

    return 0


def pump():
    with lock:
        poll_sdl()


def get(t=None):

    global event_queue

    with lock:
        poll_sdl()

        if t is None:
            rv = event_queue
            event_queue = [ ]

        elif isinstance(t, int):
            rv = get_events(( t, ))

        else:
            rv = get_events(t)

    return rv


def poll():

    with lock:
        poll_sdl()

        if event_queue:
            return event_queue.pop(0)

        return NOEVENT_EVENT


def wait():

    cdef SDL_Event evt
    cdef int result

    with lock:
        poll_sdl()

        if event_queue:
            return event_queue.pop(0)

    with nogil:
        result = SDL_WaitEvent(&evt)

    if result:
        return make_event(&evt)
    else:
        return NOEVENT_EVENT


def peek(t=None):

    with lock:
        poll_sdl()

        if t is None:
            return len(event_queue) != 0
        elif isinstance(t, int):
            return has_event(( t, ))
        else:
            return has_event(t)


def clear(t=None):

    # Clear is implemented in terms of get.
    get(t)

def get_standard_events():
    """
    Returns a list of standard events that pygame_sdl2 knows about.
    """

    return [ i for i in event_names.keys() if (i < SDL_USEREVENT) or (i > USEREVENT_MAX) ]

def event_name(t):
    try:
        return event_names[t]
    except KeyError:
        return "UNKNOWN"

def set_blocked(t=None):
    if t == None:
        for et in event_names.keys():
            SDL_EventState(et, SDL_ENABLE)
    elif isinstance(t, (int, long)):
        SDL_EventState(t, SDL_IGNORE)
    else:
        for et in t:
            SDL_EventState(et, SDL_IGNORE)

def set_allowed(t=None):
    if t == None:
        for et in event_names.keys():
            SDL_EventState(et, SDL_IGNORE)
    elif isinstance(t, (int, long)):
        SDL_EventState(t, SDL_ENABLE)
    else:
        for et in t:
            SDL_EventState(et, SDL_ENABLE)

def get_blocked(t):
    return SDL_EventState(t, SDL_QUERY) == SDL_IGNORE

def set_grab(on):
    SDL_SetWindowGrab(main_window.window, on)

def get_grab():
    return SDL_GetWindowGrab(main_window.window)

def set_mousewheel_buttons(flag):
    """
    If true (the default), the mousewheel will generate events involving
    mouse buttons 4 and 5, and mousebuttons 4 and higher will be mapped to 6 and higher.

    If false, MOUSEWHEEL events are generated, and the mousebuttons are
    not remapped.
    """

    global mousewheel_buttons
    mousewheel_buttons = flag

def get_mousewheel_buttons():
    """
    Returns the value set by mousehweel buttons,.
    """

    return mousewheel_buttons

def post(e):
    """
    Posts event object `e` to the event queue.
    """

    cdef SDL_Event event;

    if not isinstance(e, EventType):
        raise pygame_sdl2.error("event.post must be called with an Event.")

    if get_blocked(e.type):
        return

    Py_INCREF(e)

    event.type = POSTEDEVENT
    event.user.data1 = <void *> e

    SDL_PushEvent(&event)

def register(name):
    """
    Registers a unique event number and returns that number.

    `name`
        A string name for the event. This is used when calling `repr` on
        the event.
    """

    rv = SDL_RegisterEvents(1)

    event_names[rv] = name
    return rv

def copy_event_queue():
    """
    Returns a copy of the event queue. The copy cannot be used for modifying
    the event queue.
    """

    return event_queue[:]

# Usually called by display.init.
def init():
    if not SDL_WasInit(SDL_INIT_EVENTS):

        pygame_sdl2.display.sdl_main_init()

        if SDL_InitSubSystem(SDL_INIT_EVENTS):
            raise pygame_sdl2.error.error()



GL_RED_SIZE = SDL_GL_RED_SIZE
GL_GREEN_SIZE = SDL_GL_GREEN_SIZE
GL_BLUE_SIZE = SDL_GL_BLUE_SIZE
GL_ALPHA_SIZE = SDL_GL_ALPHA_SIZE
GL_BUFFER_SIZE = SDL_GL_BUFFER_SIZE
GL_DOUBLEBUFFER = SDL_GL_DOUBLEBUFFER
GL_DEPTH_SIZE = SDL_GL_DEPTH_SIZE
GL_STENCIL_SIZE = SDL_GL_STENCIL_SIZE
GL_ACCUM_RED_SIZE = SDL_GL_ACCUM_RED_SIZE
GL_ACCUM_GREEN_SIZE = SDL_GL_ACCUM_GREEN_SIZE
GL_ACCUM_BLUE_SIZE = SDL_GL_ACCUM_BLUE_SIZE
GL_ACCUM_ALPHA_SIZE = SDL_GL_ACCUM_ALPHA_SIZE
GL_STEREO = SDL_GL_STEREO
GL_MULTISAMPLEBUFFERS = SDL_GL_MULTISAMPLEBUFFERS
GL_MULTISAMPLESAMPLES = SDL_GL_MULTISAMPLESAMPLES
GL_ACCELERATED_VISUAL = SDL_GL_ACCELERATED_VISUAL
GL_RETAINED_BACKING = SDL_GL_RETAINED_BACKING
GL_CONTEXT_MAJOR_VERSION = SDL_GL_CONTEXT_MAJOR_VERSION
GL_CONTEXT_MINOR_VERSION = SDL_GL_CONTEXT_MINOR_VERSION
GL_CONTEXT_EGL = SDL_GL_CONTEXT_EGL
GL_CONTEXT_FLAGS = SDL_GL_CONTEXT_FLAGS
GL_CONTEXT_PROFILE_MASK = SDL_GL_CONTEXT_PROFILE_MASK
GL_SHARE_WITH_CURRENT_CONTEXT = SDL_GL_SHARE_WITH_CURRENT_CONTEXT
GL_FRAMEBUFFER_SRGB_CAPABLE = SDL_GL_FRAMEBUFFER_SRGB_CAPABLE
GL_CONTEXT_RELEASE_BEHAVIOR = SDL_GL_CONTEXT_RELEASE_BEHAVIOR
GL_CONTEXT_RESET_NOTIFICATION = SDL_GL_CONTEXT_RESET_NOTIFICATION
GL_CONTEXT_NO_ERROR = SDL_GL_CONTEXT_NO_ERROR



# Copyright 2014 Tom Rothamel <tom@rothamel.us>
# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
import binascii
import struct

include "color_dict.pxi"

cdef Uint32 map_color(SDL_Surface *surface, color) except? 0xaabbccdd:
    """
    Maps `color` into an RGBA color value that can be used with `surface`.
    """

    cdef Uint8 r, g, b, a

    if isinstance(color, (tuple, list, Color)) and len(color) == 4:
        r, g, b, a = color
    elif isinstance(color, (tuple, list, Color)) and len(color) == 3:
        r, g, b = color
        a = 255
    elif isinstance(color, int):
        return color
    else:
        raise TypeError("Expected a color.")

    return SDL_MapRGBA(surface.format, r, g, b, a)

cdef object get_color(Uint32 pixel, SDL_Surface *surface):
    cdef Uint8 r
    cdef Uint8 g
    cdef Uint8 b
    cdef Uint8 a

    SDL_GetRGBA(pixel, surface.format, &r, &g, &b, &a)

    return Color(r, g, b, a)

cdef to_sdl_color(color, SDL_Color *out):
    if not isinstance(color, Color):
        color = Color(color)
    out.r = color.r
    out.g = color.g
    out.b = color.b
    out.a = color.a

cdef class Color:
    cdef from_rgba(self, Uint8 r, Uint8 g, Uint8 b, Uint8 a):
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    cdef from_hex(self, c):
        # Handle short hex strings.
        if len(c) == 3 or len(c) == 4:
            c = "".join(map(lambda x: x*2, c))

        try:
            if len(c) == 6:
                r, g, b = struct.unpack('BBB', binascii.unhexlify(c))
                a = 255
            elif len(c) == 8:
                r, g, b, a = struct.unpack('BBBB', binascii.unhexlify(c))
            else:
                raise ValueError(c)
        except TypeError as e:
            raise ValueError(c)

        self.from_rgba(r, g, b, a)

    cdef from_name(self, c):
        # Remove all whitespace.
        c = "".join(c.split()).lower()

        try:
            r, g, b = colors[c]
        except KeyError as e:
            raise ValueError(c)
        self.from_rgba(r, g, b, 255)

    def __richcmp__(Color x, y, int op):
        if op == 3:
            return not (x == y)

        if isinstance(y, tuple):
            y = Color(y)
        if not isinstance(y, Color):
            return False
        if op == 2:
            return x.r == y.r and x.g == y.g and x.b == y.b and x.a == y.a

    def __cinit__(self):
        self.length = 4
        self.r = 0
        self.g = 0
        self.b = 0
        self.a = 255

    def __init__(self, *args):
        self.length = 4

        if len(args) == 1:
            c = args[0]
            if isinstance(c, basestring):
                if c.startswith('#'):
                    self.from_hex(c[1:])
                elif c.startswith('0x'):
                    self.from_hex(c[2:])
                else:
                    self.from_name(c)
            elif isinstance(c, (tuple, list, Color)):
                if len(c) == 4:
                    self.from_rgba(c[0], c[1], c[2], c[3])
                elif len(c) == 3:
                    self.from_rgba(c[0], c[1], c[2], 255)
                else:
                    raise ValueError(c)
            else:
                self.from_hex("%08x" % c)

        elif len(args) == 3:
            r, g, b = args
            self.from_rgba(r, g, b, 255)
        elif len(args) == 4:
            r, g, b, a = args
            self.from_rgba(r, g, b, a)

    def __repr__(self):
        return str((self.r, self.g, self.b, self.a))

    def __int__(self):
        packed = struct.pack('BBBB', self.r, self.g, self.b, self.a)
        return struct.unpack('>L', packed)[0]

    def __hex__(self):
        return hex(int(self))

    def __oct__(self):
        return oct(int(self))

    def __float__(self):
        return float(int(self))

    def __reduce__(self):
        d = {}
        d['rgba'] = (self.r, self.g, self.b, self.a)
        return (Color, (), d)

    def __setstate__(self, d):
        self.r, self.g, self.b, self.a = d['rgba']

    def __setitem__(self, key, val):
        if not isinstance(val, int):
            raise ValueError(val)
        if key >= self.length:
            raise IndexError(key)
        if val < 0 or val > 255:
            raise ValueError(val)

        if key == 0: self.r = val
        elif key == 1: self.g = val
        elif key == 2: self.b = val
        elif key == 3: self.a = val
        else:
            raise IndexError(key)

    def __getitem__(self, key):
        if isinstance(key, slice):
            return tuple(self)[key]
        if key >= self.length:
            raise IndexError(key)
        if key == 0: return self.r
        elif key == 1: return self.g
        elif key == 2: return self.b
        elif key == 3: return self.a
        else:
            raise IndexError(key)

    def __len__(self):
        return self.length

    def __mul__(self not None, Color rhs not None):
        # Multiplying this way doesn't make much sense,
        # but it's how pygame does it.

        r = min(255, self.r * rhs.r)
        g = min(255, self.g * rhs.g)
        b = min(255, self.b * rhs.b)
        a = min(255, self.a * rhs.a)

        return type(self)(r, g, b, a)

    def __add__(self not None, Color rhs not None):
        r = min(255, self.r + rhs.r)
        g = min(255, self.g + rhs.g)
        b = min(255, self.b + rhs.b)
        a = min(255, self.a + rhs.a)

        return type(self)(r, g, b, a)

    def __sub__(self not None, Color rhs not None):
        r = max(0, self.r - rhs.r)
        g = max(0, self.g - rhs.g)
        b = max(0, self.b - rhs.b)
        a = max(0, self.a - rhs.a)

        return type(self)(r, g, b, a)

    def __mod__(self not None, Color rhs not None):

        def mod(l, r):
            if r == 0:
                return 0

            return l % r

        r = mod(self.r, rhs.r)
        g = mod(self.g, rhs.g)
        b = mod(self.b, rhs.b)
        a = mod(self.a, rhs.a)

        return type(self)(r, g, b, a)

    def __div__(self not None, Color rhs not None):
        r = min(255, self.r / rhs.r)
        g = min(255, self.g / rhs.g)
        b = min(255, self.b / rhs.b)
        a = min(255, self.a / rhs.a)

        return type(self)(r, g, b, a)

    def __floordiv__(self not None, Color rhs not None):

        def div(a, b):
            if b == 0:
                return b

            return min(255, a // b)

        r = div(self.r, rhs.r)
        g = div(self.g, rhs.g)
        b = div(self.b, rhs.b)
        a = div(self.a, rhs.a)

        return type(self)(r, g, b, a)

    property cmy:
        def __get__(self):
            return 1 - (self.r / 255.0), 1 - (self.g / 255.0), 1 - (self.b / 255.0)

        def __set__(self, val):
            c, m, y = val
            self.r = (1 - c) * 255
            self.g = (1 - m) * 255
            self.b = (1 - y) * 255

    property hsva:
        def __get__(self):
            cdef double r = self.r / 255.0
            cdef double g = self.g / 255.0
            cdef double b = self.b / 255.0

            cdef double cmax = max(r, g, b)
            cdef double cmin = min(r, g, b)
            cdef double delta = cmax - cmin

            cdef double h, s, v, a

            if r == g == b:
                h = 0.0
                s = 0.0
            else:
                if cmax == r:
                    h = 60.0 * ((g - b) / delta % 6)
                elif cmax == g:
                    h = 60.0 * ((b - r) / delta + 2)
                else:
                    h = 60.0 * ((r - g) / delta + 4)

                if cmax == 0.0:
                    s = 0.0
                else:
                    s = delta / cmax * 100

            v = cmax * 100
            a = self.a / 255.0 * 100
            return h, s, v, a

        def __set__(self, val):
            cdef double h, s, v, a
            if len(val) == 3:
                h, s, v = val
                a = 0.0
            else:
                h, s, v, a = val

            h = h % 360.0

            # These should be in a range of [0.0, 1.0]
            s /= 100.0
            v /= 100.0
            a /= 100.0

            cdef double c = v * s
            cdef double x = c * (1 - abs((h / 60.0) % 2 - 1))
            cdef double m = v - c

            cdef double r, g, b

            if 0 <= h < 60:
                r, g, b = c, x, 0
            elif 60 <= h < 120:
                r, g, b = x, c, 0
            elif 120 <= h < 180:
                r, g, b = 0, c, x
            elif 180 <= h < 240:
                r, g, b = 0, x, c
            elif 240 <= h < 300:
                r, g, b = x, 0, c
            elif 300 <= h < 360:
                r, g, b = c, 0, x
            else:
                raise ValueError()

            self.r = int(255 * (r + m))
            self.g = int(255 * (g + m))
            self.b = int(255 * (b + m))
            self.a = int(255 * a)

    property hsla:
        def __get__(self):
            cdef double h, s, l, a
            cdef double r, g, b
            cdef double cmin, cmax, delta

            h = self.hsva[0] % 360.0

            r = self.r / 255.0
            g = self.g / 255.0
            b = self.b / 255.0

            cmin = min(r, g, b)
            cmax = max(r, g, b)
            delta = cmax - cmin

            l = (cmax + cmin) / 2.0

            if delta == 0:
                s = 0.0
            else:
                s = delta / (1 - abs(2 * l - 1))

            a = self.a / 255.0 * 100

            s = min(100.0, s * 100)
            l = min(100.0, l * 100)

            return h, s, l, a

        def __set__(self, val):
            cdef double h, s, l, a
            if len(val) == 3:
                h, s, l = val
                a = 0.0
            else:
                h, s, l, a = val

            s /= 100.0
            l /= 100.0
            a /= 100.0

            cdef double c = (1 - abs(2*l - 1)) * s
            cdef double x = c * (1 - abs((h / 60.0) % 2 - 1))
            cdef double m = l - c / 2.0

            cdef double r, g, b
            if 0 <= h < 60:
                r, g, b = c, x, 0
            elif 60 <= h < 120:
                r, g, b = x, c, 0
            elif 120 <= h < 180:
                r, g, b = 0, c, x
            elif 180 <= h < 240:
                r, g, b = 0, x, c
            elif 240 <= h < 300:
                r, g, b = x, 0, c
            elif 300 <= h < 360:
                r, g, b = c, 0, x
            else:
                raise ValueError()

            self.r = int(255 * (r + m))
            self.g = int(255 * (g + m))
            self.b = int(255 * (b + m))
            self.a = int(255 * a)

    property i1i2i3:
        def __get__(self):
            # Take the dot product as described here:
            # http://de.wikipedia.org/wiki/I1I2I3-Farbraum

            cdef double i1, i2, i3
            cdef double r, g, b

            r = self.r / 255.0
            g = self.g / 255.0
            b = self.b / 255.0

            i1 = (r + g + b) / 3.0
            i2 = (r - b) / 2.0
            i3 = (2*g - r - b) / 4.0

            return i1, i2, i3

        def __set__(self, val):
            # Dot product with the inverted matrix.

            cdef double i1, i2, i3
            cdef double r, g, b

            i1, i2, i3 = val

            r = i1 + i2 - (2.0/3.0 * i3)
            g = i1 + (4.0/3.0 * i3)
            b = i1 - i2  - (2.0/3.0 * i3)

            # Don't change alpha.
            self.r = int(r * 255)
            self.g = int(g * 255)
            self.b = int(b * 255)

    def normalize(self):
        return self.r / 255.0, self.g / 255.0, self.b / 255.0, self.a / 255.0

    def correct_gamma(self, gamma):
        m = map(lambda x: int(round(pow(x / 255.0, gamma) * 255)), tuple(self))
        c = type(self)(tuple(m))
        return c

    def set_length(self, n):
        if n > 4 or n < 1:
            raise ValueError(n)
        self.length = n



#!/usr/bin/env python

# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from __future__ import division, absolute_import, print_function

# The version of pygame_sdl2. This should also be updated in version.py
VERSION = "2.1.0"

from setuplib import android, ios, windows, cython, pymodule, setup, parse_cflags, parse_libs, find_unnecessary_gen, gen
import setuplib

import os
import platform
import shutil
import sys


def setup_env(name):
    # If PYGAME_SDL2_CC or PYGAME_SDL2_LD are in the environment, and CC or LD are not, use them.

    renpy_name = "PYGAME_SDL2_" + name
    if (renpy_name in os.environ) and (name not in os.environ):
        os.environ[name] = os.environ[renpy_name]


setup_env("CC")
setup_env("LD")
setup_env("CXX")

temporary_package_data = [ ]

if android or ios:
    sdl_libs = [ 'SDL2' ]

else:

    setuplib.package_data.extend([
        "DejaVuSans.ttf",
        "DejaVuSans.txt",
        ])

    try:
        parse_cflags([ "sh", "-c", "sdl2-config --cflags" ])
        sdl_libs = parse_libs([ "sh", "-c", "sdl2-config --libs" ])
    except:

        if not windows:
            raise

        windeps = os.path.join(os.path.dirname(__file__), "pygame_sdl2_windeps")

        if not os.path.isdir(windeps):
            raise

        sdl_libs = [ 'SDL2' ]
        setuplib.include_dirs.append(os.path.join(windeps, "include"))

        if sys.version_info[0] < 3:
            setuplib.include_dirs.append(os.path.join(windeps, "include27"))

        if platform.architecture()[0] == "32bit":
            libdir = os.path.join(windeps, "lib/x86")
        else:
            libdir = os.path.join(windeps, "lib/x64")

        setuplib.library_dirs.append(libdir)

        for i in os.listdir(libdir):
            if i.lower().endswith(".dll"):
                shutil.copy(
                    os.path.join(libdir, i),
                    os.path.join(os.path.dirname(__file__), "src", "pygame_sdl2", i),
                    )

                temporary_package_data.append(i)

        setuplib.package_data.extend(temporary_package_data)

if android:
    png = "png16"
else:
    png = "png"

pymodule("pygame_sdl2.__init__")
pymodule("pygame_sdl2.compat")
pymodule("pygame_sdl2.threads.__init__")
pymodule("pygame_sdl2.threads.Py25Queue")
pymodule("pygame_sdl2.sprite")
pymodule("pygame_sdl2.sysfont")
pymodule("pygame_sdl2.time")
pymodule("pygame_sdl2.version")

cython("pygame_sdl2.error", libs=sdl_libs)
cython("pygame_sdl2.color", libs=sdl_libs)
cython("pygame_sdl2.controller", libs=sdl_libs)
cython("pygame_sdl2.rect", libs=sdl_libs)
cython("pygame_sdl2.rwobject", libs=sdl_libs)
cython("pygame_sdl2.surface", source=[ "src/alphablit.c" ], libs=sdl_libs)
cython("pygame_sdl2.display", libs=sdl_libs)
cython("pygame_sdl2.event", libs=sdl_libs)
cython("pygame_sdl2.locals", libs=sdl_libs)
cython("pygame_sdl2.key", libs=sdl_libs)
cython("pygame_sdl2.mouse", libs=sdl_libs)
cython("pygame_sdl2.joystick", libs=sdl_libs)
cython("pygame_sdl2.power", libs=sdl_libs)
cython("pygame_sdl2.pygame_time", libs=sdl_libs)
cython("pygame_sdl2.image", source=[ "src/write_jpeg.c", "src/write_png.c" ], libs=[ 'SDL2_image', "jpeg", png ] + sdl_libs)
cython("pygame_sdl2.transform", source=[ "src/SDL2_rotozoom.c" ], libs=sdl_libs)
cython("pygame_sdl2.gfxdraw", source=[ "src/SDL_gfxPrimitives.c" ], libs=sdl_libs)
cython("pygame_sdl2.draw", libs=sdl_libs)
cython("pygame_sdl2.font", libs=['SDL2_ttf'] + sdl_libs)
cython("pygame_sdl2.mixer", libs=['SDL2_mixer'] + sdl_libs)
cython("pygame_sdl2.mixer_music", libs=['SDL2_mixer'] + sdl_libs)
cython("pygame_sdl2.scrap", libs=sdl_libs)
cython("pygame_sdl2.render", libs=['SDL2_image'] + sdl_libs)

headers = [
    "src/pygame_sdl2/pygame_sdl2.h",
    gen + "/pygame_sdl2.rwobject_api.h",
    gen + "/pygame_sdl2.surface_api.h",
    gen + "/pygame_sdl2.display_api.h",
    ]

if __name__ == "__main__":

    setup(
        "pygame_sdl2",
        VERSION,
        headers=headers,
        url="https://github.com/renpy/pygame_sdl2",
        maintainer="Tom Rothamel",
        maintainer_email="tom@rothamel.us",
        )

    find_unnecessary_gen()

    for i in temporary_package_data:
        os.unlink(os.path.join(os.path.dirname(__file__), "src", "pygame_sdl2", i))



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from sdl2_mixer cimport *
from pygame_sdl2.rwobject cimport to_rwops
from libc.string cimport memset

from pygame_sdl2.error import error

cdef Mix_Music *current_music = NULL
cdef object queued_music = None
cdef int endevent = 0

cdef void music_finished():
    global queued_music
    if queued_music:
        load(queued_music)
        play()
        queued_music = None

    cdef SDL_Event e
    if endevent != 0:
        memset(&e, 0, sizeof(SDL_Event))
        e.type = endevent
        SDL_PushEvent(&e)


def load(fi):
    global current_music

    # Free any previously loaded music.
    if current_music != NULL:
        Mix_FreeMusic(current_music)

    current_music = Mix_LoadMUS_RW(to_rwops(fi), 1)
    if current_music == NULL:
        raise error()

def play(loops=0, double start=0.0):
    Mix_FadeInMusicPos(current_music, loops, 0, start)

def rewind():
    Mix_RewindMusic()

def stop():
    Mix_HaltMusic()

def pause():
    Mix_PauseMusic()

def unpause():
    Mix_ResumeMusic()

def fadeout(time):
    Mix_FadeOutMusic(time)

def set_volume(double value):
    Mix_VolumeMusic(<int>(MIX_MAX_VOLUME * value))

def get_volume():
    return Mix_VolumeMusic(-1) / <double>MIX_MAX_VOLUME

def get_busy():
    return Mix_PlayingMusic()

def set_pos(double pos):
    Mix_SetMusicPosition(pos)

def get_pos():
    # TODO: Use a Mix_SetPostMix callback to track position.
    raise error("Not implemented.")

def queue(fi):
    Mix_HookMusicFinished(music_finished)
    if get_busy():
        global queued_music
        queued_music = fi
    else:
        load(fi)
        play()

def set_endevent(type=None):
    Mix_HookMusicFinished(music_finished)
    global endevent
    endevent = type or 0

def get_endevent():
    return endevent



# Copyright 2014 Patrick Dawson <pat@dw.is>
# Copyright 2014 Tom Rothamel <tom@rothamel.us>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from pygame_sdl2.surface cimport *
from pygame_sdl2.rect cimport to_sdl_rect

from libc.stdlib cimport calloc, free
from pygame_sdl2.locals import SRCALPHA, GL_SWAP_CONTROL
from pygame_sdl2.error import error
import pygame_sdl2

import warnings
import os

# True if we are on ios.
ios = ("PYGAME_IOS" in os.environ)

# This inits SDL proper, and should be called by the other init methods.

# A map from a PYGAME_SDL2 hint to what it was set to.
_pygame_hints = { }

def hint(hint, value, priority=1):

    if str(hint).startswith("PYGAME_SDL2"):
        _pygame_hints[str(hint)] = str(value)
        return

    if not isinstance(hint, bytes):
        hint = hint.encode("utf-8")

    if not isinstance(value, bytes):
        value = value.encode("utf-8")

    SDL_SetHintWithPriority(hint, value, priority)

def _get_hint(hint, default):
    hint = str(hint)

    if hint in _pygame_hints:
        return _pygame_hints[hint]

    if hint in os.environ:
        return os.environ[hint]

    return default


main_done = False

def sdl_main_init():
    global main_done

    if main_done:
        return

    SDL_SetMainReady()

    if SDL_Init(0):
        raise error()

    main_done = True

# True if init has been called without quit being called.
init_done = False

@pygame_sdl2.register_init
def init():

    if init_done:
        return

    sdl_main_init()

    hint("SDL_VIDEO_MAC_FULLSCREEN_SPACES", "0")

    if SDL_InitSubSystem(SDL_INIT_VIDEO):
        raise error()

    pygame_sdl2.event.init()

    global init_done
    init_done = True



@pygame_sdl2.register_quit
def quit(): # @ReservedAssignment

    global init_done
    global main_window

    if main_window:
        main_window.destroy()
        main_window = None

    init_done = False

def get_init():
    return init_done


# The window that is used by the various module globals.
main_window = None

# Have we shown the first window?
_shown_first_window = False

def _before_first_window():
    global _shown_first_window

    if _shown_first_window:
        return

    _shown_first_window = True

    # If we're on android, we have to close the splash window before opening
    # our window.
    try:
        import androidembed
        androidembed.close_window()
    except ImportError:
        pass



cdef class Window:
    def __init__(self, title, resolution=(0, 0), flags=0, depth=0, pos=(SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED)):

        if not isinstance(title, bytes):
            title = title.encode("utf-8")

        _before_first_window()

        self.create_flags = flags

        # If we do not get the AVOID_GL hint, we always create a GL-compatible
        # window. This lets us change the OPENGL flag later on.
        if int(_get_hint("PYGAME_SDL2_AVOID_GL", "0")):
            gl_flag = 0
        else:
            gl_flag = SDL_WINDOW_OPENGL

        self.window = SDL_CreateWindow(
            title,
            pos[0], pos[1],
            resolution[0], resolution[1], flags | gl_flag)

        if not self.window:
            raise error()

        # From here on, the window exists. So we have to call self.destroy if
        # an exception occurs.

        try:

            if flags & SDL_WINDOW_OPENGL:

                self.gl_context = SDL_GL_CreateContext(self.window)

                if self.gl_context == NULL:
                    raise error()

                if not ios:
                    # Try setting the swap interval - first positive, then negated
                    # to deal with the case where the negative interval isn't
                    # supported. Then give up and carry on.
                    if SDL_GL_SetSwapInterval(default_swap_control):
                        SDL_GL_SetSwapInterval(-default_swap_control)


            self.create_surface()

        except:
            self.destroy()
            raise

    def create_surface(self):
        """
        Creates the surface associated with this window.
        """

        cdef int w, h
        SDL_GetWindowSize(self.window, &w, &h)

        if self.gl_context:

            # For now, make this the size of the window so get_size() works.
            # TODO: Make this a bit less wasteful of memory, even if it means
            # we lie about the actual size of the pixel array.
            self.surface = Surface((w, h), SRCALPHA, 32)

        else:

            self.window_surface = SDL_GetWindowSurface(self.window)

            # If the surface is 32-bit, we can use it directly. Otherwise,
            # we need to create a 32-bit proxy surface.
            if self.window_surface.format.BitsPerPixel == 32:

                self.surface = Surface(())
                self.surface.surface = self.window_surface
                self.surface.owns_surface = False
                self.surface.window_surface = True

            else:
                self.surface = Surface((w, h), 0, 32)

        self.surface.get_window_flags = self.get_window_flags


    def destroy(self):
        """
        This should be called before the window is deleted.
        """

        if self.gl_context != NULL:
            SDL_GL_DeleteContext(self.gl_context)

        if self.surface:

            # Break the cycle that prevents refcounting from collecting this
            # object.
            self.surface.get_window_flags = None

            # Necessary to collect the GL surface, doesn't hurt the window surface.
            self.surface = None

        SDL_DestroyWindow(self.window)

    def resize(self, size, opengl=False, fullscreen=None):
        """
        Resizes the window to `size`, which must be a width, height tuple. If opengl
        is true, adds an OpenGL context, if it's missing. Otherwise, removes the
        opengl context if present.
        """

        if fullscreen is None:
            fullscreen = SDL_GetWindowFlags(self.window) & SDL_WINDOW_FULLSCREEN_DESKTOP

        # Prevents a loop between the surface and this object.
        self.surface.get_window_flags = None

        if self.gl_context and not opengl:
            SDL_GL_DeleteContext(self.gl_context)
            self.gl_context = NULL

        cdef int cur_width = 0
        cdef int cur_height = 0

        if fullscreen:
            if SDL_SetWindowFullscreen(self.window, SDL_WINDOW_FULLSCREEN_DESKTOP):
                fullscreen = False

        if not fullscreen:
            SDL_SetWindowFullscreen(self.window, 0)

            width, height = size

            SDL_GetWindowSize(self.window, &cur_width, &cur_height)

            if (cur_width != width) or (cur_height != height):
                SDL_SetWindowSize(self.window, width, height)

        # Create a missing GL context.
        if opengl and not self.gl_context:
            self.gl_context = SDL_GL_CreateContext(self.window)

            if self.gl_context == NULL:
                raise error()

        self.create_surface()

    def recreate_gl_context(self):
        """
        Check to see if the GL context was lost, and re-create it if it was.
        """

        if <unsigned int> SDL_GL_GetCurrentContext():
            return False

        self.gl_context = SDL_GL_CreateContext(self.window)

        if self.gl_context == NULL:
            raise error()

        return True

    def get_window_flags(self):
        rv = SDL_GetWindowFlags(self.window)

        if self.gl_context:
            rv |= SDL_WINDOW_OPENGL
        else:
            rv &= ~SDL_WINDOW_OPENGL

        return rv

    def proxy_window_surface(self):
        SDL_UpperBlit(self.surface.surface, NULL, self.window_surface, NULL)

    def flip(self):
        cdef const char *err

        if self.gl_context != NULL:
            with nogil:
                SDL_ClearError();

                SDL_GL_SwapWindow(self.window)

                err = SDL_GetError()

            if err[0]:
                raise error(err)

        else:

            if self.surface.surface != self.window_surface:
                self.proxy_window_surface()

            with nogil:
                SDL_UpdateWindowSurface(self.window)

    def get_surface(self):
        return self.surface

    def update(self, rectangles=None):

        cdef SDL_Rect *rects
        cdef int count = 0

        if rectangles is None:
            self.flip()
            return

        if self.surface.surface != self.window_surface:
            self.proxy_window_surface()

        if not isinstance(rectangles, list):
            rectangles = [ rectangles ]

        rects = <SDL_Rect *> calloc(len(rectangles), sizeof(SDL_Rect))
        if rects == NULL:
            raise MemoryError("Couldn't allocate rectangles.")

        try:

            for i in rectangles:
                if i is None:
                    continue

                to_sdl_rect(i, &rects[count])
                count += 1

            SDL_UpdateWindowSurfaceRects(self.window, rects, count)

        finally:
            free(rects)

    def get_wm_info(self):
        return { }

    def get_active(self):
        if SDL_GetWindowFlags(self.window) & (SDL_WINDOW_HIDDEN | SDL_WINDOW_MINIMIZED):
            return False
        else:
            return True

    def iconify(self):
        SDL_MinimizeWindow(self.window)
        return True

    def toggle_fullscreen(self):
        if SDL_GetWindowFlags(self.window) & (SDL_WINDOW_FULLSCREEN_DESKTOP):
            if SDL_SetWindowFullscreen(self.window, 0):
                raise error()
        else:
            if SDL_SetWindowFullscreen(self.window, SDL_WINDOW_FULLSCREEN_DESKTOP):
                raise error()

        return True

    def set_gamma(self, red, green=None, blue=None):
        if green is None:
            green = red
        if blue is None:
            blue = red

        cdef Uint16 red_gamma[256]
        cdef Uint16 green_gamma[256]
        cdef Uint16 blue_gamma[256]

        SDL_CalculateGammaRamp(red, red_gamma)
        SDL_CalculateGammaRamp(green, green_gamma)
        SDL_CalculateGammaRamp(blue, blue_gamma)

        if SDL_SetWindowGammaRamp(self.window, red_gamma, green_gamma, blue_gamma):
            return False

        return True

    def set_gamma_ramp(self, red, green, blue):

        cdef Uint16 red_gamma[256]
        cdef Uint16 green_gamma[256]
        cdef Uint16 blue_gamma[256]

        for i in range(256):
            red_gamma[i] = red[i]
            green_gamma[i] = green[i]
            blue_gamma[i] = blue[i]

        if SDL_SetWindowGammaRamp(self.window, red_gamma, green_gamma, blue_gamma):
            return False

        return True

    def set_icon(self, Surface surface):
        SDL_SetWindowIcon(self.window, surface.surface)

    def set_caption(self, title):

        if not isinstance(title, bytes):
            title = title.encode("utf-8")

        SDL_SetWindowTitle(self.window, title)

    def get_drawable_size(self):
        cdef int w, h

        SDL_GL_GetDrawableSize(self.window, &w, &h)
        return w, h


    def get_size(self):
        cdef int w, h

        SDL_GetWindowSize(self.window, &w, &h)
        return w, h

    def restore(self):
        SDL_RestoreWindow(self.window)

    def maximize(self):
        SDL_MaximizeWindow(self.window)

    def minimize(self):
        SDL_MinimizeWindow(self.window)


# The icon that's used for new windows.
default_icon = None

# The title that's used for new windows.
default_title = "pygame window"

# The default gl_swap_control
default_swap_control = 1

def set_mode(resolution=(0, 0), flags=0, depth=0, pos=(SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED)):
    global main_window

    RESIZE_FLAGS = SDL_WINDOW_OPENGL | SDL_WINDOW_FULLSCREEN_DESKTOP

    if main_window:

        if (flags & ~RESIZE_FLAGS) == (main_window.create_flags & ~RESIZE_FLAGS):
            main_window.resize(resolution, flags & SDL_WINDOW_OPENGL, flags & SDL_WINDOW_FULLSCREEN_DESKTOP)
            return main_window.surface

        else:
            main_window.destroy()

    main_window = Window(default_title, resolution, flags, depth, pos=pos)

    if default_icon is not None:
        main_window.set_icon(default_icon)

    return main_window.surface

def get_surface():
    if main_window is None:
        return None

    return main_window.get_surface()

def get_window():
    """
    Returns the Window created by set_mode, or None if no such window
    exists.
    """

    return main_window

def flip():
    if main_window:
        main_window.flip()

def update(rectangles=None):
    if main_window:
        main_window.update(rectangles)

def get_driver():
    cdef const char *driver = SDL_GetCurrentVideoDriver()

    if driver == NULL:
        raise error()

    return driver

class Info(object):

    def __init__(self):
        cdef SDL_DisplayMode dm
        cdef SDL_PixelFormat *format

        if SDL_GetCurrentDisplayMode(0, &dm):
            raise error()

        format = SDL_AllocFormat(dm.format)
        if format == NULL:
            raise error()

        self.bitsize = format.BitsPerPixel
        self.bytesize = format.BytesPerPixel

        self.masks = (
            format.Rmask,
            format.Gmask,
            format.Bmask,
            format.Amask,
            )

        self.shifts = (
            format.Rshift,
            format.Gshift,
            format.Bshift,
            format.Ashift,
            )

        self.losses = (
            format.Rloss,
            format.Gloss,
            format.Bloss,
            format.Aloss,
            )

        SDL_FreeFormat(format)

        if main_window:
            self.current_w, self.current_h = main_window.surface.get_size()
        else:

            self.current_w = dm.w
            self.current_h = dm.h

        self.refresh_rate = dm.refresh_rate

        # The rest of these are just guesses.
        self.hw = False
        self.wm = True
        self.video_mem = 256 * 1024 * 1024

        self.blit_hw = False
        self.blit_hw_CC = False
        self.blit_hw_A = False

        self.blit_sw = False
        self.blit_sw_CC = False
        self.blit_sw_A = False

    def __repr__(self):
        return "<Info({!r})>".format(self.__dict__)


def get_wm_info():
    if main_window:
        return main_window.get_wm_info()

    return {}


def get_num_video_displays():
    """
    Returns the number of video displays connected to the system.
    """

    rv = SDL_GetNumVideoDisplays()

    if rv < 0:
        raise error()

    return rv


def list_modes(depth=0, flags=0, display=0):
    """
    Returns a list of possible display modes for the display `display`.

    `depth` and `flags` are ignored.
    """

    cdef int num_modes, i
    cdef SDL_DisplayMode mode

    rv = [ ]

    num_modes = SDL_GetNumDisplayModes(display)
    if num_modes < 0:
        raise error()

    for 0 <= i < num_modes:
        if SDL_GetDisplayMode(display, i, &mode) == 0:
            t = (mode.w, mode.h)
            if t not in rv:
                rv.append(t)

    return rv


def mode_ok(size, flags=0, depth=0):
    """
    Returns true if size is in the result of list_modes().
    """

    return tuple(size) in list_modes()

def gl_reset_attributes():
    SDL_GL_ResetAttributes()

def gl_set_attribute(flag, value):

    if flag == GL_SWAP_CONTROL:
        if ios:
            return

        global default_swap_control

        # Try setting the swap interval - first positive, then negated
        # to deal with the case where the negative interval isn't
        # supported. Then give up and carry on.
        if SDL_GL_SetSwapInterval(value):
            SDL_GL_SetSwapInterval(-value)

        default_swap_control = value
        return

    if SDL_GL_SetAttribute(flag, value):
        raise error()

def gl_get_attribute(flag):
    cdef int rv

    if flag == GL_SWAP_CONTROL:
        return SDL_GL_GetSwapInterval()

    if SDL_GL_GetAttribute(flag, &rv):
        raise error()

    return rv

def gl_load_library(path):
    if path is None:
        if SDL_GL_LoadLibrary(NULL):
            raise error()
    else:
        if SDL_GL_LoadLibrary(path):
            raise error()

def gl_unload_library():
    SDL_GL_UnloadLibrary()

def get_active():
    if main_window:
        return main_window.get_active()
    return False

def iconify():
    if main_window:
        return main_window.iconify()

    return False

def toggle_fullscreen():
    if main_window:
        return main_window.toggle_fullscreen()

    return True

def set_gamma(red, green=None, blue=None):
    if main_window:
        return main_window.set_gamma(red, green, blue)
    return False

def set_gamma_ramp(red, green, blue):
    if main_window:
        return main_window.set_gamma_ramp(red, green, blue)
    return False

def set_icon(surface):
    global default_icon

    default_icon = surface.copy()

    if main_window is not None:
        main_window.set_icon(default_icon)

def set_caption(title, icontitle = None):
    global default_title

    default_title = title

    if main_window:
        main_window.set_caption(default_title)

def get_caption():
    return default_title

def get_drawable_size():
    if main_window:
        return main_window.get_drawable_size()
    return None

def get_size():
    if main_window:
        return main_window.get_size()
    return None

def get_num_video_displays():
    rv = SDL_GetNumVideoDisplays()
    if rv < 0:
        raise error()

    return rv

def get_display_bounds(index):
    cdef SDL_Rect rect
    rv = SDL_GetDisplayBounds(index, &rect)

    return (rect.x, rect.y, rect.w, rect.h)

def set_screensaver(state):
    """
    Sets the screenslaver to `state`.
    """

    if state:
        SDL_EnableScreenSaver()
    else:
        SDL_DisableScreenSaver()

def get_platform():
    return SDL_GetPlatform()

cdef api SDL_Window *PyWindow_AsWindow(window):
    """
    Returns a pointer to the SDL_Window corresponding to `window`. If `window`
    is None, a pointer to the main window is returned. NULL is returned if
    there is no main window.
    """

    if window is None:
        window = main_window

    if window is None:
        return NULL

    return (<Window> window).window



# Automatically generated by generate_event_list.py
# Do not edit!

NOEVENT = SDL_FIRSTEVENT
QUIT = SDL_QUIT
APP_TERMINATING = SDL_APP_TERMINATING
APP_LOWMEMORY = SDL_APP_LOWMEMORY
APP_WILLENTERBACKGROUND = SDL_APP_WILLENTERBACKGROUND
APP_DIDENTERBACKGROUND = SDL_APP_DIDENTERBACKGROUND
APP_WILLENTERFOREGROUND = SDL_APP_WILLENTERFOREGROUND
APP_DIDENTERFOREGROUND = SDL_APP_DIDENTERFOREGROUND
DISPLAYEVENT = SDL_DISPLAYEVENT
WINDOWEVENT = SDL_WINDOWEVENT
SYSWMEVENT = SDL_SYSWMEVENT
KEYDOWN = SDL_KEYDOWN
KEYUP = SDL_KEYUP
TEXTEDITING = SDL_TEXTEDITING
TEXTINPUT = SDL_TEXTINPUT
KEYMAPCHANGED = SDL_KEYMAPCHANGED
MOUSEMOTION = SDL_MOUSEMOTION
MOUSEBUTTONDOWN = SDL_MOUSEBUTTONDOWN
MOUSEBUTTONUP = SDL_MOUSEBUTTONUP
MOUSEWHEEL = SDL_MOUSEWHEEL
JOYAXISMOTION = SDL_JOYAXISMOTION
JOYBALLMOTION = SDL_JOYBALLMOTION
JOYHATMOTION = SDL_JOYHATMOTION
JOYBUTTONDOWN = SDL_JOYBUTTONDOWN
JOYBUTTONUP = SDL_JOYBUTTONUP
JOYDEVICEADDED = SDL_JOYDEVICEADDED
JOYDEVICEREMOVED = SDL_JOYDEVICEREMOVED
CONTROLLERAXISMOTION = SDL_CONTROLLERAXISMOTION
CONTROLLERBUTTONDOWN = SDL_CONTROLLERBUTTONDOWN
CONTROLLERBUTTONUP = SDL_CONTROLLERBUTTONUP
CONTROLLERDEVICEADDED = SDL_CONTROLLERDEVICEADDED
CONTROLLERDEVICEREMOVED = SDL_CONTROLLERDEVICEREMOVED
CONTROLLERDEVICEREMAPPED = SDL_CONTROLLERDEVICEREMAPPED
FINGERDOWN = SDL_FINGERDOWN
FINGERUP = SDL_FINGERUP
FINGERMOTION = SDL_FINGERMOTION
DOLLARGESTURE = SDL_DOLLARGESTURE
DOLLARRECORD = SDL_DOLLARRECORD
MULTIGESTURE = SDL_MULTIGESTURE
CLIPBOARDUPDATE = SDL_CLIPBOARDUPDATE
DROPFILE = SDL_DROPFILE
DROPTEXT = SDL_DROPTEXT
DROPBEGIN = SDL_DROPBEGIN
DROPCOMPLETE = SDL_DROPCOMPLETE
AUDIODEVICEADDED = SDL_AUDIODEVICEADDED
AUDIODEVICEREMOVED = SDL_AUDIODEVICEREMOVED
SENSORUPDATE = SDL_SENSORUPDATE
RENDER_TARGETS_RESET = SDL_RENDER_TARGETS_RESET
RENDER_DEVICE_RESET = SDL_RENDER_DEVICE_RESET
USEREVENT = SDL_USEREVENT
LASTEVENT = SDL_LASTEVENT



# Copyright 2014 Patrick Dawson <pat@dw.is>
#
# This software is provided 'as-is', without any express or implied
# warranty.  In no event will the authors be held liable for any damages
# arising from the use of this software.
#
# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:
#
# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.

from sdl2 cimport *
from sdl2_mixer cimport *
from pygame_sdl2.rwobject cimport to_rwops
from libc.string cimport memset

import sys
import threading
from pygame_sdl2.error import error

import pygame_sdl2.mixer_music as music
import pygame_sdl2

cdef object preinit_args = None
cdef object output_spec = None

cdef dict channel_events = {}
cdef dict channel_queued = {}
cdef dict current_sounds = {}

# The lock protects channel_queued and current_sounds.
_lock = threading.Lock()

def _play_current(int channel):
    """
    Caled by channel_callback to play the next sound. This has to be called
    from a different thread, as the channel callback isn't allowed to call
    MIX functions.
    """

    cdef Sound next_sound

    with _lock:
        next_sound = channel_queued[channel]
        current_sounds[channel] = next_sound
        channel_queued[channel] = None

    if next_sound:
        with nogil:
            Mix_PlayChannelTimed(channel, next_sound.chunk, 0, -1)


cdef void channel_callback(int channel) with gil:

    cdef int etype = 0
    cdef SDL_Event e

    etype = channel_events.get(channel, 0)
    if etype != 0:
        memset(&e, 0, sizeof(SDL_Event))
        e.type = etype
        SDL_PushEvent(&e)

    with _lock:
        next_sound = channel_queued.get(channel)
        if next_sound:
            threading.Thread(target=_play_current, args=(channel,)).start()

# A list of errors that occured during mixer initialization.
errors = [ ]

@pygame_sdl2.register_init
def init(frequency=22050, size=MIX_DEFAULT_FORMAT, channels=2, buffer=4096):
    if get_init() is not None:
        return

    for flag in (MIX_INIT_FLAC, MIX_INIT_MP3, MIX_INIT_OGG):

        if Mix_Init(flag) != flag:
            errors.append("{}\n".format(SDL_GetError()))

    if preinit_args:
        frequency, size, channels, buffer = preinit_args

    if Mix_OpenAudio(frequency, size, channels, buffer) != 0:
        raise error()

    global output_spec
    output_spec = get_init()

    Mix_ChannelFinished(channel_callback)

def pre_init(frequency=22050, size=MIX_DEFAULT_FORMAT, channels=2, buffersize=4096):
    global preinit_args
    preinit_args = (frequency, size, channels, buffersize)

@pygame_sdl2.register_quit
def quit(): # @ReservedAssignment
    Mix_CloseAudio()
    Mix_Quit()

def get_init():
    cdef int frequency
    cdef Uint16 format
    cdef int channels

    if Mix_QuerySpec(&frequency, &format, &channels) == 0:
        return None
    else:
        return frequency, format, channels

def stop():
    with nogil:
        Mix_HaltChannel(-1)

def pause():
    with nogil:
        Mix_Pause(-1)

def unpause():
    with nogil:
        Mix_Resume(-1)

def fadeout(time):
    cdef int ms = time
    with nogil:
        Mix_FadeOutChannel(-1, ms)

def set_num_channels(count):
    Mix_AllocateChannels(count)

def get_num_channels():
    return Mix_AllocateChannels(-1)

def set_reserved(count):
    Mix_ReserveChannels(count)

def find_channel(force=False):
    cdef int chan
    chan = Mix_GroupAvailable(-1)
    if chan == -1:
        if not force:
            return None
        chan = Mix_GroupOldest(-1)
        if chan == -1:
            raise error()
    return Channel(chan)

def get_busy():
    return Mix_GroupNewer(-1) != -1


cdef class Sound:
    cdef Mix_Chunk *chunk

    def __cinit__(self):
        self.chunk = NULL

    def __dealloc__(self):
        if self.chunk:
            Mix_FreeChunk(self.chunk)

    def __init__(self, fi):
        self.chunk = Mix_LoadWAV_RW(to_rwops(fi), 1)
        if self.chunk == NULL:
            raise error()

    def play(self, loops=0, maxtime=-1, fade_ms=0):
        cdef int cid
        cdef int _loops = loops
        cdef int _maxtime = maxtime
        cdef int _fade_ms = fade_ms

        with nogil:
            if _fade_ms != 0:
                cid = Mix_FadeInChannelTimed(-1, self.chunk, _loops, _fade_ms, _maxtime)
            else:
                cid = Mix_PlayChannelTimed(-1, self.chunk, _loops, _maxtime)

        if cid == -1:
            raise error()
        return Channel(cid)

    def stop(self):
        cdef int i = 0
        while i < Mix_AllocateChannels(-1):
            if Mix_GetChunk(i) == self.chunk:
                with nogil:
                    Mix_HaltChannel(i)
            i += 1

    def pause(self):
        cdef int i = 0
        while i < Mix_AllocateChannels(-1):
            if Mix_GetChunk(i) == self.chunk:
                with nogil:
                    Mix_Pause(i)
            i += 1

    def unpause(self):
        cdef int i = 0
        while i < Mix_AllocateChannels(-1):
            if Mix_GetChunk(i) == self.chunk:
                with nogil:
                    Mix_Resume(i)
            i += 1

    def fadeout(self, time):
        cdef int i = 0
        cdef int ms = time
        while i < Mix_AllocateChannels(-1):
            if Mix_GetChunk(i) == self.chunk:
                with nogil:
                    Mix_FadeOutChannel(i, ms)
            i += 1

    def set_volume(self, value):
        Mix_VolumeChunk(self.chunk, MIX_MAX_VOLUME * value)

    def get_volume(self):
        return Mix_VolumeChunk(self.chunk, -1)

    def get_num_channels(self):
        cdef int i = 0
        cdef int n = 0
        while i < Mix_AllocateChannels(-1):
            if Mix_GetChunk(i) == self.chunk:
                n += 1
            i += 1
        return n

    def get_length(self):
        # TODO: Adjust for actual format, rather than assuming 16-bit.
        return <double>self.chunk.alen / output_spec[0] / 2 / output_spec[2]

    def get_raw(self):
        # return self.chunk.abuf
        raise error("Not implemented.")


cdef class Channel(object):
    cdef int cid

    def __init__(self, cid):
        self.cid = cid

    def play(self, Sound sound not None, loops=0, maxtime=-1, fade_ms=0):
        cdef int _loops = loops
        cdef int _maxtime = maxtime
        cdef int _fade_ms = fade_ms

        with nogil:
            if _fade_ms != 0:
                cid = Mix_FadeInChannelTimed(self.cid, sound.chunk, _loops, _fade_ms, _maxtime)
            else:
                cid = Mix_PlayChannelTimed(self.cid, sound.chunk, _loops, _maxtime)

        if cid == -1:
            raise error()

        with _lock:
            current_sounds[self.cid] = sound

    def stop(self):
        with nogil:
            Mix_HaltChannel(self.cid)

    def pause(self):
        with nogil:
            Mix_Pause(self.cid)

    def unpause(self):
        with nogil:
            Mix_Resume(self.cid)

    def fadeout(self, time):
        cdef int ms = time
        with nogil:
            Mix_FadeOutChannel(self.cid, ms)

    def set_volume(self, volume):
        Mix_Volume(self.cid, int(MIX_MAX_VOLUME * volume))

    def get_volume(self):
        cdef int vol = Mix_Volume(self.cid, -1)
        return vol / <double>MIX_MAX_VOLUME

    def get_busy(self):
        return Mix_Playing(self.cid) != 0

    def get_sound(self):
        with _lock:
            return current_sounds.get(self.cid)

    def queue(self, Sound sound):
        if self.get_busy():
            with _lock:
                channel_queued[self.cid] = sound
        else:
            self.play(sound)

    def get_queue(self):
        with _lock:
            return channel_queued.get(self.cid)

    def set_endevent(self, type=None):
        channel_events[self.cid] = type or 0

    def get_endevent(self):
        return channel_events.get(self.cid, 0)
